# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple
from datetime import datetime
from datetime import timedelta
from uuid import UUID
import logging
import struct

from hwp5.dataio import Struct
from hwp5.dataio import Flags
from hwp5.dataio import N_ARRAY
from hwp5.dataio import ARRAY
from hwp5.dataio import BYTE
from hwp5.dataio import UINT16
from hwp5.dataio import UINT32
from hwp5.dataio import INT32
from hwp5.bintype import read_type


logger = logging.getLogger(__name__)


vt_types = dict()


def PropertyType(code):

    def decorator(cls):
        cls.code = code
        vt_types[code] = cls
        return cls

    return decorator


@PropertyType(code=0x0003)
class VT_I4(object):

    @classmethod
    def read_value(cls, context, f):
        return read_type(INT32, context, f)


@PropertyType(code=0x001F)
class VT_LPWSTR(object):

    @classmethod
    def read_value(cls, context, f):
        length = read_type(UINT32, context, f)
        data = f.read(length * 2)
        return data.decode('utf-16le')[:-1]  # remove null character


@PropertyType(code=0x0040)
class VT_FILETIME(object):

    @classmethod
    def read_value(cls, context, f):
        lword = read_type(UINT32, context, f)
        hword = read_type(UINT32, context, f)
        value = hword << 32 | lword
        value = FILETIME(value)
        return value


class FILETIME(object):
    __slots__ = ('value', )

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.datetime)

    @property
    def datetime(self):
        return (
            datetime(1601, 1, 1, 0, 0, 0) +
            timedelta(microseconds=self.value / 10)
        )


PropertyIdentifier = namedtuple('PropertyIdentifier', [
    'id',
    'label',
])


PID_DICTIONARY = PropertyIdentifier(
    id=0x00000000,
    label='PID_DICTIONARY',
)
PID_CODEPAGE = PropertyIdentifier(
    id=0x00000001,
    label='PID_CODEPAGE',
)
PID_LOCALE = PropertyIdentifier(
    id=0x80000000,
    label='PID_LOCALE',
)
PID_BEHAVIOR = PropertyIdentifier(
    id=0x80000003,
    label='PID_BEHAVIOR',
)
PIDSI_TITLE = PropertyIdentifier(
    id=0x02,
    label='PIDSI_TITLE'
)
PIDSI_SUBJECT = PropertyIdentifier(
    id=0x03,
    label='PIDSI_SUBJECT'
)
PIDSI_AUTHOR = PropertyIdentifier(
    id=0x04,
    label='PIDSI_AUTHOR'
)
PIDSI_KEYWORDS = PropertyIdentifier(
    id=0x05,
    label='PIDSI_KEYWORDS'
)
PIDSI_COMMENTS = PropertyIdentifier(
    id=0x06,
    label='PIDSI_COMMENTS'
)
PIDSI_TEMPLATE = PropertyIdentifier(
    id=0x07,
    label='PIDSI_TEMPLATE'
)
PIDSI_LASTAUTHOR = PropertyIdentifier(
    id=0x08,
    label='PIDSI_LASTAUTHOR'
)
PIDSI_REVNUMBER = PropertyIdentifier(
    id=0x09,
    label='PIDSI_REVNUMBER'
)
PIDSI_EDITTIME = PropertyIdentifier(
    id=0x0a,
    label='PIDSI_EDITTIME'
)
PIDSI_LASTPRINTED = PropertyIdentifier(
    id=0x0b,
    label='PIDSI_LASTPRINTED'
)
PIDSI_CREATE_DTM = PropertyIdentifier(
    id=0x0c,
    label='PIDSI_CREATE_DTM'
)
PIDSI_LASTSAVE_DTM = PropertyIdentifier(
    id=0x0d,
    label='PIDSI_LASTSAVE_DTM'
)
PIDSI_PAGECOUNT = PropertyIdentifier(
    id=0x0e,
    label='PIDSI_PAGECOUNT'
)
PIDSI_WORDCOUNT = PropertyIdentifier(
    id=0x0f,
    label='PIDSI_WORDCOUNT'
)
PIDSI_CHARCOUNT = PropertyIdentifier(
    id=0x10,
    label='PIDSI_CHARCOUNT'
)
PIDSI_THUMBNAIL = PropertyIdentifier(
    id=0x11,
    label='PIDSI_THUMBNAIL'
)
PIDSI_APPNAME = PropertyIdentifier(
    id=0x12,
    label='PIDSI_APPNAME'
)
PIDSI_SECURITY = PropertyIdentifier(
    id=0x13,
    label='PIDSI_SECURITY'
)


RESERVED_PROPERTIES = (
    PID_DICTIONARY,
    PID_CODEPAGE,
    PID_LOCALE,
    PID_BEHAVIOR,
)


SUMMARY_INFORMATION_PROPERTIES = (
    PIDSI_TITLE,
    PIDSI_SUBJECT,
    PIDSI_AUTHOR,
    PIDSI_KEYWORDS,
    PIDSI_COMMENTS,
    PIDSI_TEMPLATE,
    PIDSI_LASTAUTHOR,
    PIDSI_REVNUMBER,
    PIDSI_EDITTIME,
    PIDSI_LASTPRINTED,
    PIDSI_CREATE_DTM,
    PIDSI_LASTSAVE_DTM,
    PIDSI_PAGECOUNT,
    PIDSI_WORDCOUNT,
    PIDSI_CHARCOUNT,
    PIDSI_THUMBNAIL,
    PIDSI_APPNAME,
    PIDSI_SECURITY,
)


class Property(object):

    def __init__(self, desc, idLabel, type, value):
        self.desc = desc
        self.idLabel = idLabel
        self.type = type
        self.value = value

    @property
    def id(self):
        return self.desc.id


class PropertyDesc(Struct):

    def __init__(self, id, offset):
        self.id = id
        self.offset = offset

    @classmethod
    def fromDict(cls, d):
        return cls(id=d['id'], offset=d['offset'])

    def attributes():
        yield UINT32, 'id'
        yield UINT32, 'offset'  # offset from section start
    attributes = staticmethod(attributes)


class PropertyReader(object):

    def __init__(self, propsetDesc, propDesc, idLabel, codepage,
                 displayName=None):
        self.propsetDesc = propsetDesc
        self.propDesc = propDesc
        self.idLabel = idLabel
        self.codepage = codepage
        self.displayName = displayName

    def read(self, f):
        f.seek(self.propsetDesc.offset + self.propDesc.offset)

        context = {}
        propType = read_type(TypedPropertyValue, context, f)
        propType = TypedPropertyValue.fromDict(propType)
        vt_type = vt_types[propType.code]
        propValue = vt_type.read_value(context, f)

        return Property(
            desc=self.propDesc,
            idLabel=self.idLabel,
            type=propType,
            value=propValue,
        )


class TypedPropertyValue(Struct):
    '''
    [MS-OLEPS] 2.15 TypedPropertyValue
    '''

    def __init__(self, code):
        self.code = code

    @classmethod
    def fromDict(cls, d):
        return cls(code=d['type'].code)

    TypeFlags = Flags(UINT32,
                      0, 16, 'code')

    def attributes(cls):
        yield cls.TypeFlags, 'type'
    attributes = classmethod(attributes)

    @property
    def vt_type(self):
        try:
            return vt_types[self.code]
        except KeyError:
            return None


class DictionaryEntry(Struct):
    '''
    [MS-OLEPS] 2.16 DictionaryEntry
    '''

    def __init__(self, id, name):
        self.id = id
        self.name = name

    @classmethod
    def fromDict(cls, d):
        return cls(
            id=d['id'],
            name=nullterminated_string(d['name']),
        )

    def attributes():
        from hwp5.dataio import N_ARRAY
        from hwp5.dataio import BYTE
        yield UINT32, 'id'
        yield N_ARRAY(UINT32, BYTE), 'name'
    attributes = staticmethod(attributes)


class Dictionary(Struct):
    '''
    [MS-OLEPS] 2.17 Dictionary
    '''

    def __init__(self, entries):
        self.entries = entries

    @classmethod
    def fromDict(cls, d):
        entries = tuple(
            DictionaryEntry.fromDict(entry)
            for entry in d['entries']
        )
        return cls(entries=entries)

    def attributes():
        from hwp5.dataio import N_ARRAY
        yield N_ARRAY(UINT32, DictionaryEntry), 'entries'
    attributes = staticmethod(attributes)

    def get(self, id, defvalue=None):
        for entry in self.entries:
            if id == entry.id:
                return entry.name
        return defvalue


class DictionaryReader(object):

    def __init__(self, propsetDesc, propDesc, idLabel, codepage):
        self.propsetDesc = propsetDesc
        self.propDesc = propDesc
        self.idLabel = idLabel
        self.codepage = codepage

    def read(self, f):
        propsetDesc = self.propsetDesc
        propDesc = self.propDesc
        idLabel = self.idLabel

        f.seek(propsetDesc.offset + propDesc.offset)
        context = {}
        propType = None
        propValue = read_type(Dictionary, context, f)
        propValue = Dictionary.fromDict(propValue)
        return Property(
            desc=propDesc,
            idLabel=idLabel,
            type=propType,
            value=propValue,
        )


class PropertySet(object):
    '''
    [MS-OLEPS] 2.20 PropertySet
    '''

    def __init__(self, desc, header, properties):
        self.desc = desc
        self.header = header
        self.properties = properties

    @property
    def fmtid(self):
        return self.desc.fmtid

    def __getitem__(self, propertyIdentifier):
        for property in self.properties:
            if property.id == propertyIdentifier.id:
                return property.value
        raise KeyError(propertyIdentifier)


class PropertySetHeader(Struct):

    def __init__(self, bytesize, propDescList):
        self.bytesize = bytesize,
        self.propDescList = propDescList

    @classmethod
    def fromDict(cls, d):
        return cls(
            bytesize=d['bytesize'],
            propDescList=tuple(
                PropertyDesc.fromDict(
                    propDesc
                )
                for propDesc in d['propDescList']
            ),
        )

    def attributes():
        from hwp5.dataio import N_ARRAY
        yield UINT32, 'bytesize'
        yield N_ARRAY(UINT32, PropertyDesc), 'propDescList'
    attributes = staticmethod(attributes)


class PropertySetDesc(Struct):

    def __init__(self, fmtid, offset):
        self.fmtid = fmtid
        self.offset = offset

    def attributes():
        yield ARRAY(BYTE, 16), 'fmtid'
        yield UINT32, 'offset'
    attributes = staticmethod(attributes)

    @classmethod
    def fromDict(cls, d):
        return cls(
            fmtid=uuid_from_bytes_tuple(d['fmtid']),
            offset=d['offset'],
        )


class PropertySetStreamHeader(Struct):

    def __init__(self, byteOrder, version, systemIdentifier, clsid,
                 propsetDescList):
        self.byteOrder = byteOrder
        self.version = version
        self.systemIdentifier = systemIdentifier
        self.clsid = clsid
        self.propsetDescList = propsetDescList

    @classmethod
    def fromDict(cls, d):
        return cls(
            byteOrder=d['byteOrder'],
            version=d['version'],
            systemIdentifier=d['systemIdentifier'],
            clsid=uuid_from_bytes_tuple(d['clsid']),
            propsetDescList=tuple(
                PropertySetDesc.fromDict(
                    propsetDesc
                )
                for propsetDesc in d['propsetDescList']
            )
        )

    def attributes():
        yield UINT16, 'byteOrder'
        yield UINT16, 'version'
        yield UINT32, 'systemIdentifier'
        yield ARRAY(BYTE, 16), 'clsid'
        yield N_ARRAY(UINT32, PropertySetDesc), 'propsetDescList'
    attributes = staticmethod(attributes)


class PropertySetStream(object):
    '''
    [MS-OLEPS] 2.21 PropertySetStream
    '''

    def __init__(self, header, propertysets):
        self.header = header
        self.propertysets = propertysets

    @property
    def byteOrder(self):
        return self.header.byteOrder

    @property
    def version(self):
        return self.header.version

    @property
    def systemIdentifier(self):
        return self.header.systemIdentifier

    @property
    def clsid(self):
        return self.header.clsid


class PropertySetFormat(object):

    def __init__(self, fmtid, propertyIdentifiers):
        self.fmtid = fmtid
        self.propertyIdentifiers = propertyIdentifiers

    @property
    def idLabels(self):
        return {
            p.id: p.label
            for p in self.propertyIdentifiers
        }


class PropertySetStreamReader(object):

    def __init__(self, propertySetFormats):
        self.propertySetFormats = {
            propsetFormat.fmtid: propsetFormat
            for propsetFormat in propertySetFormats
        }

    def read(self, f):
        context = {}
        streamHeader = read_type(PropertySetStreamHeader, context, f)
        streamHeader = PropertySetStreamHeader.fromDict(streamHeader)
        propertysetList = list()
        for propsetDesc in streamHeader.propsetDescList:
            f.seek(propsetDesc.offset)
            propsetHeader = read_type(PropertySetHeader, context, f)
            propsetHeader = PropertySetHeader.fromDict(
                propsetHeader,
            )
            try:
                propsetFormat = self.propertySetFormats[propsetDesc.fmtid]
            except KeyError:
                idLabels = {}
            else:
                idLabels = propsetFormat.idLabels

            properties = []
            propDescMap = {
                propDesc.id: propDesc
                for propDesc in propsetHeader.propDescList
            }

            propDesc = propDescMap.pop(PID_CODEPAGE.id, None)
            if propDesc is not None:
                idLabel = idLabels.get(propDesc.id)
                propReader = PropertyReader(
                    propsetDesc=propsetDesc,
                    propDesc=propDesc,
                    idLabel=idLabel,
                    codepage=None,
                    displayName=None,
                )
                prop = propReader.read(f)
                properties.append(prop)

                codepage = prop.value
            else:
                codepage = None

            propDesc = propDescMap.pop(PID_DICTIONARY.id, None)
            if propDesc is not None:
                idLabel = idLabels.get(propDesc.id)
                propReader = DictionaryReader(
                    propsetDesc,
                    propDesc,
                    idLabel,
                    codepage,
                )
                prop = propReader.read(f)
                properties.append(prop)

                dictionary = prop.value
            else:
                dictionary = None

            for propDesc in propDescMap.values():
                idLabel = idLabels.get(propDesc.id)
                displayName = dictionary.get(propDesc.id, None)
                propReader = PropertyReader(
                    propsetDesc=propsetDesc,
                    propDesc=propDesc,
                    idLabel=idLabel,
                    codepage=codepage,
                    displayName=displayName,
                )
                prop = propReader.read(f)
                properties.append(prop)

            propertyset = PropertySet(
                desc=propsetDesc,
                header=propsetHeader,
                properties=properties,
            )
            propertysetList.append(propertyset)

        return PropertySetStream(
            header=streamHeader,
            propertysets=propertysetList,
        )


class PropertySetStreamTextFormatter(object):

    def formatTextLines(self, stream):
        yield '- ByteOrder: 0x%x' % stream.byteOrder
        yield '- Version: %d' % stream.version
        yield '- SystemIdentifier: 0x%08x' % stream.systemIdentifier
        yield '- CLSID: %s' % stream.clsid
        yield ''

        for propertyset in stream.propertysets:
            title = 'Property Set {}'.format(
                propertyset.fmtid,
            )
            yield '- {:08x}: {}'.format(
                propertyset.desc.offset,
                title,
            )
            yield '            {}'.format(
                '-' * len(title)
            )

            properties = sorted(
                propertyset.properties,
                key=lambda property: property.desc.offset,
            )
            for property in properties:
                if property.id == PID_DICTIONARY.id:
                    yield '- {:08x}: {}(=0x{:08x}):'.format(
                        propertyset.desc.offset + property.desc.offset,
                        property.idLabel if property.idLabel is not None
                        else '',
                        property.id,
                    )
                    for entry in property.value.entries:
                        yield '            - {}: {}'.format(
                            entry.id,
                            entry.name,
                        )
                else:
                    yield '- {:08x}: {}(=0x{:08x}): {}'.format(
                        propertyset.desc.offset + property.desc.offset,
                        property.idLabel if property.idLabel is not None
                        else '',
                        property.id,
                        property.value
                    )


def uuid_from_bytes_tuple(t):
    fmt = 'B' * len(t)
    fmt = '<' + fmt
    bytes_le = struct.pack(fmt, *t)
    return UUID(bytes_le=bytes_le)


def nullterminated_string(bs):
    return ''.join(chr(x) for x in bs)[:-1]



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Transform an HWPv5 file into an XML.

.. note::

   This command is experimental. Its output format is subject to change at any
   time.

Usage::

    hwp5proc xml [--embedbin]
                 [--no-xml-decl]
                 [--output=<file>]
                 [--format=<format>]
                 [--no-validate-wellformed]
                 [--loglevel=<loglevel>] [--logfile=<logfile>]
                 <hwp5file>
    hwp5proc xml --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --embedbin           Embed BinData/* streams in the output XML.
       --no-xml-decl        Don't output <?xml ... ?> XML declaration.
       --output=<file>      Output filename.

    <hwp5file>              HWPv5 files (*.hwp)
    <format>                "flat", "nested" (default: "nested")

Example::

    $ hwp5proc xml samples/sample-5017.hwp > sample-5017.xml
    $ xmllint --format sample-5017.xml

With ``--embedbin`` option, you can embed base64-encoded ``BinData/*`` files in
the output XML.

Example::

    $ hwp5proc xml --embedbin samples/sample-5017.hwp > sample-5017.xml
    $ xmllint --format sample-5017.xml

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from functools import partial
import logging

from ..utils import make_open_dest_file
from ..utils import wrap_open_dest_for_tty
from ..utils import wrap_open_dest
from ..utils import pager
from ..utils import syntaxhighlight
from ..utils import xmllint
from ..xmldump_flat import xmldump_flat
from ..xmlmodel import Hwp5File


logger = logging.getLogger(__name__)


def xmldump_nested(hwp5file, output, embedbin=False, xml_declaration=True):
    dump = hwp5file.xmlevents(embedbin=embedbin).dump
    dump = partial(dump, xml_declaration=xml_declaration)
    dump(output)


def main(args):
    ''' Transform <hwp5file> into an XML.
    '''

    fmt = args['--format'] or 'nested'
    if fmt == 'flat':
        xmldump = partial(
            xmldump_flat,
            xml_declaration=not args['--no-xml-decl']
        )
    elif fmt == 'nested':
        xmldump = partial(
            xmldump_nested,
            xml_declaration=not args['--no-xml-decl'],
            embedbin=args['--embedbin'],
        )

    open_dest = make_open_dest_file(args['--output'])
    open_dest = wrap_open_dest_for_tty(open_dest, [
        pager(),
        syntaxhighlight('application/xml'),
    ] + ([
        xmllint(format=True),
    ] if not args['--no-validate-wellformed'] else []))
    open_dest = wrap_open_dest(open_dest, [
        xmllint(encode='utf-8'),
        xmllint(c14n=True),
    ] if not args['--no-validate-wellformed'] else [])

    hwp5file = Hwp5File(args['<hwp5file>'])
    with open_dest() as output:
        xmldump(hwp5file, output)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_CHAR_SHAPE
from hwp5.dataio import StructType
from hwp5.dataio import Struct
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import INT32
from hwp5.dataio import INT8
from hwp5.dataio import UINT8
from hwp5.dataio import WORD
from hwp5.binmodel._shared import COLORREF


def LanguageStruct(name, basetype):
    ''' 표 29 글꼴에 대한 언어 '''
    def attributes():
        for lang in ('ko', 'en', 'cn', 'jp', 'other', 'symbol', 'user'):
            yield basetype, lang
    attributes = staticmethod(attributes)
    return StructType(name, (Struct,), dict(basetype=basetype,
                                            attributes=attributes))


class ShadowSpace(Struct):
    def attributes():
        yield INT8, 'x'
        yield INT8, 'y'
    attributes = staticmethod(attributes)


class CharShape(RecordModel):
    ''' 4.1.6. 글자 모양 '''

    tagid = HWPTAG_CHAR_SHAPE

    # 표 30 글자 모양 속성
    Underline = Enum(NONE=0, UNDERLINE=1, LINE_THROUGH=2, OVERLINE=3)
    UnderlineStyle = Enum(SOLID=0, DASHED=1, DOTTED=2, DASH_DOT=3,
                          DASH_DOT_DOT=4, LONG_DASHED=5, LARGE_DOTTED=6,
                          DOUBLE=7, LOWER_WEIGHTED=8, UPPER_WEIGHTED=9,
                          MIDDLE_WEIGHTED=10)
    Flags = Flags(UINT32,
                  0, 'italic',
                  1, 'bold',
                  2, 3, Underline, 'underline',
                  4, 7, UnderlineStyle, 'underline_style',
                  8, 10, 'outline',
                  11, 13, 'shadow')

    def attributes(cls):
        ''' 표 28 글자 모양 '''
        yield LanguageStruct('FontFace', WORD), 'font_face',
        yield (LanguageStruct('LetterWidthExpansion', UINT8),
               'letter_width_expansion')
        yield LanguageStruct('LetterSpacing', INT8), 'letter_spacing'
        yield LanguageStruct('RelativeSize', INT8), 'relative_size'
        yield LanguageStruct('Position', INT8), 'position'
        yield INT32, 'basesize',
        yield cls.Flags, 'charshapeflags',
        yield ShadowSpace, 'shadow_space'
        yield COLORREF, 'text_color',
        yield COLORREF, 'underline_color',
        yield COLORREF, 'shade_color',
        yield COLORREF, 'shadow_color',
        # yield UINT16, 'borderfill_id',        # DIFFSPEC
        # yield COLORREF, 'strikeoutColor',    # DIFFSPEC
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


def create_service(context, name, *args):
    sm = context.ServiceManager
    if len(args) > 0:
        return sm.createInstanceWithArgumentsAndContext(name, args, context)
    else:
        return sm.createInstanceWithContext(name, context)


class Namespace(object):
    def __init__(self, dotted_name):
        self.dotted_name = dotted_name

    def __getattr__(self, name):
        return Namespace(self.dotted_name + '.' + name)

    def __call__(self, context, *args):
        return create_service(context, self.dotted_name, *args)

    def bind(self, context):
        return ContextBoundNamespace(self, context)


class ContextBoundNamespace(object):

    def __init__(self, namespace, context):
        self.namespace = namespace
        self.context = context

    def __getattr__(self, name):
        obj = getattr(self.namespace, name, None)
        if isinstance(obj, Namespace):
            return obj.bind(self.context)
        return obj

    def __call__(self, *args):
        return self.namespace(self.context, *args)

    def __iter__(self):
        context = self.context
        sm = context.ServiceManager
        prefix = self.dotted_name + '.'
        for name in sm.AvailableServiceNames:
            if name.startswith(prefix):
                basename = name[len(prefix):]
                if basename.find('.') == -1:
                    yield basename


css = Namespace('com.sun.star')



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from six import with_metaclass

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PARA_CHAR_SHAPE
from hwp5.dataio import ArrayType
from hwp5.dataio import X_ARRAY
from hwp5.dataio import ARRAY
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.binmodel._shared import ref_parent_member


class ParaCharShape(RecordModel):
    ''' 4.2.3. 문단의 글자 모양 '''
    tagid = HWPTAG_PARA_CHAR_SHAPE

    def attributes():
        ''' 표 56 문단의 글자 모양 '''
        yield dict(name='charshapes',
                   type=X_ARRAY(ARRAY(UINT32, 2),
                                ref_parent_member('charshapes')))
    attributes = staticmethod(attributes)


class ParaCharShapeList(with_metaclass(ArrayType, list)):

    itemtype = ARRAY(UINT16, 2)

    def read(cls, f, context):
        bytes = f.read()
        return cls.decode(bytes, context)
    read = classmethod(read)

    def decode(payload, context=None):
        import struct
        fmt = 'II'
        unitsize = struct.calcsize('<' + fmt)
        unitcount = len(payload) / unitsize
        values = struct.unpack('<' + (fmt * unitcount), payload)
        return list(tuple(values[i * 2:i * 2 + 2])
                    for i in range(0, unitcount))
    decode = staticmethod(decode)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Decode distribute doc stream.

Usage::

    hwp5proc diststream
    hwp5proc diststream sha1 [--raw]
    hwp5proc diststream key [--raw]
    hwp5proc diststream [--loglevel=<loglevel>] [--logfile=<logfile>]
    hwp5proc diststream --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

Example::

    $ hwp5proc cat --ole samples/viewtext.hwp ViewText/Section0
      | tee Section0.zraw.aes128ecb | hwp5proc diststream | tee Section0.zraw
      | hwp5proc rawunz > Section0

    $ hwp5proc diststream sha1 < Section0.zraw.aes128ecb
    $ echo -n '12345' | sha1sum

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from binascii import b2a_hex
from binascii import a2b_hex
import logging
import shutil
import sys

from ..distdoc import decode
from ..distdoc import decode_head_to_sha1
from ..distdoc import decode_head_to_key
from ..recordstream import read_record


PY2 = sys.version_info.major == 2
logger = logging.getLogger(__name__)


def main(args):
    if PY2:
        input_fp = sys.stdin
        output_fp = sys.stdout
    else:
        input_fp = sys.stdin.buffer
        output_fp = sys.stdout.buffer

    if args['sha1']:
        head = read_record(input_fp, 0)
        sha1ucs16le = decode_head_to_sha1(head['payload'])
        sha1 = a2b_hex(sha1ucs16le.decode('utf-16le'))
        if not args['--raw']:
            sha1 = b2a_hex(sha1)
        output_fp.write(sha1)
    elif args['key']:
        head = read_record(input_fp, 0)
        key = decode_head_to_key(head['payload'])
        if not args['--raw']:
            key = b2a_hex(key)
        output_fp.write(key)
    else:
        result = decode(input_fp)
        shutil.copyfileobj(result, output_fp)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

from six import with_metaclass

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PARA_TEXT
from hwp5.dataio import ArrayType
from hwp5.binmodel.controlchar import ControlChar


class ParaTextChunks(with_metaclass(ArrayType, list)):

    def read(cls, f):
        bytes = f.read()
        return [x for x in cls.parse_chunks(bytes)]
    read = classmethod(read)

    def parse_chunks(bytes):
        from hwp5.dataio import decode_utf16le_with_hypua
        size = len(bytes)
        idx = 0
        while idx < size:
            ctrlpos, ctrlpos_end = ControlChar.find(bytes, idx)
            if idx < ctrlpos:
                text = decode_utf16le_with_hypua(bytes[idx:ctrlpos])
                yield (idx // 2, ctrlpos // 2), text
            if ctrlpos < ctrlpos_end:
                cch = ControlChar.decode(bytes[ctrlpos:ctrlpos_end])
                yield (ctrlpos // 2, ctrlpos_end // 2), cch
            idx = ctrlpos_end
    parse_chunks = staticmethod(parse_chunks)


class ParaText(RecordModel):
    ''' 4.2.2. 문단의 텍스트 '''
    tagid = HWPTAG_PARA_TEXT

    def attributes():
        yield ParaTextChunks, 'chunks'
    attributes = staticmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Find record models with specified predicates.

Usage::

    hwp5proc find [--model=<model-name> | --tag=<hwptag>]
                  [--incomplete] [--dump] [--format=<format>]
                  [--loglevel=<loglevel>] [--logfile=<logfile>]
                  (--from-stdin | <hwp5files>...)
    hwp5proc find --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --from-stdin         get filenames fro stdin

       --model=<model-name> filter with record model name
       --tag=<hwptag>       filter with record HWPTAG
       --incomplete         filter with incompletely parsed content

       --format=<format>    record output format
                            %(filename)s %(stream)s %(seqno)s %(type)s
       --dump               dump record

    <hwp5files>...          HWPv5 files (*.hwp)

Example: Find paragraphs::

    $ hwp5proc find --model=Paragraph samples/*.hwp
    $ hwp5proc find --tag=HWPTAG_PARA_TEXT samples/*.hwp
    $ hwp5proc find --tag=66 samples/*.hwp

Example: Find and dump records of ``HWPTAG_LIST_HEADER`` which is parsed
incompletely::

    $ hwp5proc find --tag=HWPTAG_LIST_HEADER --incomplete --dump samples/*.hwp

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from functools import partial
import itertools
import sys

from ..binmodel import Hwp5File
from ..binmodel import model_to_json
from ..bintype import log_events
from ..dataio import ParseError
from ..tagids import tagnames
from . import logger


PY2 = sys.version_info.major == 2
if PY2:
    ifilter = itertools.ifilter
    imap = itertools.imap
else:
    ifilter = filter
    imap = map


def main(args):
    filenames = filenames_from_args(args)

    conditions = list(conditions_from_args(args))
    filter_conditions = partial(
        ifilter, lambda m: all(condition(m) for condition in conditions)
    )

    print_model = printer_from_args(args)

    for filename in filenames:
        try:
            models = hwp5file_models(filename)
            models = filter_conditions(models)
            for model in models:
                print_model(model)
        except ParseError as e:
            logger.error('---- On processing %s:', filename)
            e.print_to_logger(logger)


def filenames_from_args(args):
    if args['--from-stdin']:
        return filenames_from_stdin(args)
    return args['<hwp5files>']


def filenames_from_stdin(args):
    return imap(lambda line: line[:-1], sys.stdin)


def conditions_from_args(args):

    if args['--model']:
        def with_model_name(model):
            return args['--model'] == model['type'].__name__
        yield with_model_name

    if args['--tag']:
        tag = args['--tag']
        try:
            tag = int(tag)
        except ValueError:
            pass
        else:
            tag = tagnames[tag]

        def with_tag(model):
            return model['tagname'] == tag
        yield with_tag

    if args['--incomplete']:
        def with_incomplete(model):
            return 'unparsed' in model
        yield with_incomplete


def hwp5file_models(filename):
    hwp5file = Hwp5File(filename)
    for model in flat_models(hwp5file):
        model['filename'] = filename
        yield model


def flat_models(hwp5file, **kwargs):
    for model in hwp5file.docinfo.models(**kwargs):
        model['stream'] = 'DocInfo'
        yield model

    for section in hwp5file.bodytext:
        for model in hwp5file.bodytext[section].models(**kwargs):
            model['stream'] = 'BodyText/' + section
            yield model


def printer_from_args(args):

    if args['--format']:
        fmt = args['--format']
    else:
        fmt = '%(filename)s %(stream)s %(seqno)s %(tagname)s %(type)s'

    dump = args['--dump']

    def print_model(model):
        printable_model = dict(model, type=model['type'].__name__)
        print(fmt % printable_model)
        if dump:
            print(model_to_json(model, sort_keys=True, indent=2))

            def print_log(fmt, *args):
                print(fmt % args)
            list(log_events(model['binevents'], print_log))
    return print_model



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''HWP5 viewer (Experimental, gtk only)

Usage::

    hwp5view [options] <hwp5file>
    hwp5view -h | --help
    hwp5view --version

Options::

    -h --help           Show this screen
    --version           Show version
    --loglevel=<level>  Set log level.
    --logfile=<file>    Set log file.
'''
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from contextlib import closing
from contextlib import contextmanager
from tempfile import mkdtemp
import gettext
import io
import logging
import os.path
import shutil
import sys
import urllib

from docopt import docopt

from hwp5 import __version__ as version
from hwp5.proc import rest_to_docopt
from hwp5.proc import init_logger
from hwp5.xmlmodel import Hwp5File
from hwp5.hwp5html import HTMLTransform


PY3 = sys.version_info.major == 3
logger = logging.getLogger(__name__)
locale_dir = os.path.join(os.path.dirname(__file__), '..', 'locale')
locale_dir = os.path.abspath(locale_dir)
t = gettext.translation('hwp5view', locale_dir, fallback=True)
if PY3:
    _ = t.gettext
else:
    _ = t.ugettext


def main():
    doc = rest_to_docopt(__doc__)
    args = docopt(doc, version=version)
    init_logger(args)

    runner = runner_factory()

    with make_temporary_directory() as out_dir:
        with hwp5html(args['<hwp5file>'], out_dir) as index_path:
            runner(args, index_path, out_dir)


def runner_factory():
    try:
        return runner_factory_gi()
    except ImportError:
        pass

    try:
        return runner_factory_pyside()
    except ImportError:
        pass

    raise NotImplementedError(
        'Neither gi.repository.WebKit nor pyside is found'
    )


def runner_factory_gi():
    from gi.repository import Gtk
    from gi.repository import WebKit

    def runner(args, index_path, out_dir):
        base_uri = fspath2url(out_dir) + '/'
        # index_uri = fspath2url(index_path)
        with io.open(index_path, 'rb') as f:
            content = f.read()

        view = WebKit.WebView()
        # view.load_uri(index_uri)
        view.load_string(content, 'text/html', 'utf-8', base_uri)

        def on_load(webview, webframe):
            script = ('window.location.href = "dimension:" '
                      '+ document.body.scrollWidth + "x"'
                      '+ document.body.scrollHeight')
            webview.execute_script(script)

        MIN_WIDTH = 300
        MIN_HEIGHT = 400
        MAX_WIDTH = 1024
        MAX_HEIGHT = 800

        view.connect('load-finished', on_load)

        def on_navigation_requested(webview, frame, req, data=None):
            uri = req.get_uri()
            scheme, path = uri.split(':', 1)
            if scheme == 'dimension':
                width, height = path.split('x', 1)
                width = int(width)
                height = int(height)
                width = min(width, MAX_WIDTH)
                height = min(height, MAX_HEIGHT)
                width = max(width, MIN_WIDTH)
                height = max(height, MIN_HEIGHT)
                window.resize(width + 4, height)
                return True
            return False

        view.connect('navigation-requested', on_navigation_requested)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.add(view)

        vbox = Gtk.VBox()
        vbox.pack_start(scrolled_window, expand=True, fill=True, padding=0)

        window = Gtk.Window()
        window.add(vbox)
        window.connect('delete-event', Gtk.main_quit)
        window.set_default_size(600, 800)
        window.show_all()

        Gtk.main()

    return runner


def runner_factory_pyside():
    from PySide.QtCore import QUrl
    from PySide.QtGui import QApplication
    from PySide.QtGui import QMainWindow
    from PySide.QtWebKit import QWebView

    class MainWindow(QMainWindow):
        pass

    def runner(args, index_path, out_dir):
        app = QApplication(sys.argv)

        frame = MainWindow()
        frame.setWindowTitle('hwp5view')
        frame.setMinimumWidth(400)

        url = fspath2url(index_path)
        url = QUrl(url)
        view = QWebView(frame)

        logger.info('Loading...')
        view.load(url)

        @view.loadFinished.connect
        def onLoadFinished():
            frame.show()

        frame.setCentralWidget(view)

        app.exec_()

    return runner


@contextmanager
def make_temporary_directory(*args, **kwargs):
    path = mkdtemp(*args, **kwargs)
    try:
        logger.warning('temporary directory for contents: %s', path)
        yield path
    finally:
        shutil.rmtree(path)


@contextmanager
def hwp5html(filename, out_dir):
    with closing(Hwp5File(filename)) as hwp5file:
        HTMLTransform().transform_hwp5_to_dir(hwp5file, out_dir)
        yield os.path.join(out_dir, 'index.xhtml')


def fspath2url(path):
    path = os.path.abspath(path)
    return 'file://' + urllib.pathname2url(path)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_FACE_NAME
from hwp5.dataio import Flags
from hwp5.dataio import Enum
from hwp5.dataio import BSTR
from hwp5.dataio import BYTE
from hwp5.dataio import Struct


class AlternateFont(Struct):
    def attributes():
        yield BYTE, 'kind'
        yield BSTR, 'name'
    attributes = staticmethod(attributes)


class Panose1(Struct):
    ''' 표 17 글꼴 유형 정보 '''

    FamilyType = Enum('any', 'no_fit', 'text_display', 'script', 'decorative',
                      'pictorial')

    SerifStyle = Enum('any', 'no_fit', 'cove', 'obtuse_cove', 'square_cove',
                      'obtuse_square_cove', 'square', 'thin', 'bone',
                      'exaggerated', 'triangle', 'normal_sans', 'obtuse_sans',
                      'perp_sans', 'flared', 'rounded')

    Weight = Enum('any', 'no_fit', 'very_light', 'light', 'thin', 'book',
                  'medium', 'demi', 'bold', 'heavy', 'black', 'nord')

    Proportion = Enum('any', 'no_fit', 'old_style', 'modern', 'even_width',
                      'expanded', 'condensed', 'very_expanded',
                      'very_condensed', 'monospaced')

    Contrast = Enum('any', 'no_fit', 'none', 'very_low', 'low', 'medium_low',
                    'medium', 'medium_high', 'high', 'very_high')

    StrokeVariation = Enum('any', 'no_fit', 'gradual_diag', 'gradual_tran',
                           'gradual_vert', 'gradual_horz', 'rapid_vert',
                           'rapid_horz', 'instant_vert')

    ArmStyle = Enum('any', 'no_fit', 'straight_horz', 'straight_wedge',
                    'straight_vert', 'straight_single_serif',
                    'straight_double_serif', 'bent_horz', 'bent_wedge',
                    'bent_vert', 'bent_single_serif', 'bent_double_serif')

    Letterform = Enum('any', 'no_fit', 'normal_contact', 'normal_weighted',
                      'normal_boxed', 'normal_flattened', 'normal_rounded',
                      'normal_off_center', 'normal_square', 'oblique_contact',
                      'oblique_weighted', 'oblique_boxed', 'oblique_flattened',
                      'oblique_rounded', 'oblique_off_center',
                      'oblique_square')

    Midline = Enum('any', 'no_fit', 'standard_trimmed', 'standard_pointed',
                   'standard_serifed', 'high_trimmed', 'high_pointed',
                   'high_serifed', 'constant_trimmed', 'constant_pointed',
                   'constant_serifed', 'low_trimmed', 'low_pointed',
                   'low_serifed')

    XHeight = Enum('any', 'no_fit', 'constant_small', 'constant_std',
                   'constant_large', 'ducking_small', 'ducking_std',
                   'ducking_large')

    def attributes():
        yield BYTE, 'family_type',
        yield BYTE, 'serif_style',
        yield BYTE, 'weight',
        yield BYTE, 'proportion',
        yield BYTE, 'contrast',
        yield BYTE, 'stroke_variation',
        yield BYTE, 'arm_style',
        yield BYTE, 'letterform',
        yield BYTE, 'midline',
        yield BYTE, 'x_height',
    attributes = staticmethod(attributes)


class FaceName(RecordModel):
    ''' 4.1.4. 글꼴 '''

    tagid = HWPTAG_FACE_NAME

    # 표 16 대체 글꼴 유형
    FontFileType = Enum(UNKNOWN=0, TTF=1, HFT=2)

    # 표 15 글꼴 속성
    Flags = Flags(BYTE,
                  0, 1, FontFileType, 'font_file_type',
                  5, 'default',
                  6, 'metric',
                  7, 'alternate')

    def attributes(cls):
        ''' 표 14 글꼴 '''
        yield cls.Flags, 'flags'
        yield BSTR, 'name'

        def has_alternate(context, values):
            ''' flags.alternate == 1 '''
            return values['flags'].alternate

        def has_metric(context, values):
            ''' flags.metric == 1 '''
            return values['flags'].metric

        def has_default(context, values):
            ''' flags.default == 1 '''
            return values['flags'].default

        yield dict(type=AlternateFont, name='alternate_font',
                   condition=has_alternate)
        yield dict(type=Panose1, name='panose1', condition=has_metric)
        yield dict(type=BSTR, name='default_font', condition=has_default)
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from six import with_metaclass

from hwp5.binmodel._shared import RecordModelType
from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_CTRL_HEADER
from hwp5.binmodel.controlchar import CHID


control_models = dict()


class ControlType(RecordModelType):

    def __new__(mcs, name, bases, attrs):
        cls = RecordModelType.__new__(mcs, name, bases, attrs)
        if 'chid' in attrs:
            chid = attrs['chid']
            assert chid not in control_models
            control_models[chid] = cls
        return cls


class Control(with_metaclass(ControlType, RecordModel)):
    ''' 4.2.6. 컨트롤 헤더 '''

    tagid = HWPTAG_CTRL_HEADER

    def attributes():
        yield CHID, 'chid'
    attributes = staticmethod(attributes)

    extension_types = control_models

    def get_extension_key(cls, context, model):
        ''' chid '''
        return model['content']['chid']
    get_extension_key = classmethod(get_extension_key)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015,2019 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from binascii import b2a_hex
from functools import partial
from subprocess import CalledProcessError
from subprocess import Popen
import logging
import os
import subprocess
import tempfile

from . import _lxml
from . import _uno
from . import gir_gsf
from . import javax_transform
from . import jython_poifs
from . import olefileio
from . import xmllint
from . import xsltproc


logger = logging.getLogger(__name__)


def get_xslt():
    if javax_transform.is_enabled():
        return javax_transform.xslt
    if _lxml.is_enabled():
        return _lxml.xslt
    if xsltproc.is_enabled():
        return xsltproc.xslt
    if _uno.is_enabled():
        return _uno.xslt


def get_xslt_compile():
    modules = [
        javax_transform,
        _lxml,
        xsltproc,
        _uno
    ]
    for module in modules:
        if module.is_enabled():
            xslt_compile = getattr(module, 'xslt_compile', None)
            if xslt_compile:
                return xslt_compile
            xslt = getattr(module, 'xslt', None)
            if xslt:
                def xslt_compile(xsl_path):
                    return partial(xslt, xsl_path)


def get_relaxng():
    if _lxml.is_enabled():
        return _lxml.relaxng
    if xmllint.is_enabled():
        return xmllint.relaxng


def get_relaxng_compile():
    modules = [
        _lxml,
        xmllint,
    ]
    for module in modules:
        if module.is_enabled():
            relaxng_compile = getattr(module, 'relaxng_compile', None)
            if relaxng_compile:
                return relaxng_compile
            relaxng = getattr(module, 'relaxng', None)
            if relaxng:
                def relaxng_compile(rng_path):
                    return partial(relaxng, rng_path)


def get_olestorage_class():
    if jython_poifs.is_enabled():
        return jython_poifs.OleStorage
    if olefileio.is_enabled():
        return olefileio.OleStorage
    if _uno.is_enabled():
        return _uno.OleStorage
    if gir_gsf.is_enabled():
        return gir_gsf.OleStorage


def get_aes128ecb_decrypt():
    try:
        return get_aes128ecb_decrypt_cryptography()
    except Exception:
        pass

    try:
        return get_aes128ecb_decrypt_javax()
    except Exception:
        pass

    try:
        return get_aes128ecb_decrypt_openssl()
    except Exception:
        pass

    raise NotImplementedError('aes128ecb_decrypt')


def get_aes128ecb_decrypt_cryptography():
    from cryptography.hazmat.primitives.ciphers import Cipher
    from cryptography.hazmat.primitives.ciphers import algorithms
    from cryptography.hazmat.primitives.ciphers import modes
    from cryptography.hazmat.backends import default_backend

    def decrypt(key, ciphertext):
        backend = default_backend()
        cipher = Cipher(algorithms.AES(key), modes.ECB(), backend=backend)
        decryptor = cipher.decryptor()
        return decryptor.update(ciphertext) + decryptor.finalize()

    return decrypt


def get_aes128ecb_decrypt_javax():
    from javax.crypto import Cipher
    from javax.crypto.spec import SecretKeySpec

    def decrypt(key, ciphertext):
        secretkey = SecretKeySpec(key, 'AES')
        cipher = Cipher.getInstance('AES/ECB/NoPadding')
        cipher.init(Cipher.DECRYPT_MODE, secretkey)
        decrypted = cipher.doFinal(ciphertext)
        return decrypted.tostring()

    return decrypt


def get_aes128ecb_decrypt_openssl():
    if not openssl_reachable():
        raise NotImplementedError()

    def decrypt(key, ciphertext):
        fd, name = tempfile.mkstemp()
        fp = os.fdopen(fd, 'wb')
        try:
            fp.write(ciphertext)
        finally:
            fp.close()

        args = [
            'openssl',
            'enc',
            '-d',
            '-in',
            name,
            '-aes-128-ecb',
            '-K',
            b2a_hex(key),
            '-nopad',
        ]
        try:
            p = Popen(args, stdout=subprocess.PIPE)
            try:
                return p.stdout.read()
            finally:
                p.wait()
                p.stdout.close()
        finally:
            os.unlink(name)

    return decrypt


def openssl_reachable():
    args = ['openssl', 'version']
    try:
        subprocess.check_output(args)
    except OSError:
        return False
    except CalledProcessError:
        return False
    except Exception as e:
        logger.exception(e)
        return False
    else:
        return True



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from ..errors import InvalidOleStorageError
from ..utils import cached_property


def is_enabled():
    try:
        import olefile  # noqa
    except Exception:
        pass
    else:
        return True

    try:
        import OleFileIO_PL  # noqa
    except ImportError:
        pass
    else:
        return True

    return False


def import_isOleFile():
    try:
        from olefile import isOleFile
    except ImportError:
        pass
    else:
        return isOleFile

    try:
        from OleFileIO_PL import isOleFile
    except ImportError:
        pass
    else:
        return isOleFile


def import_OleFileIO():
    try:
        from olefile import OleFileIO
    except ImportError:
        pass
    else:
        return OleFileIO

    try:
        from OleFileIO_PL import OleFileIO
    except ImportError:
        pass
    else:
        return OleFileIO


class OleStorageItem(object):

    def __init__(self, olefile, path, parent=None):
        self.olefile = olefile
        self.path = path  # path DOES NOT end with '/'

    def get_name(self):
        if self.path == '':
            return None
        segments = self.path.split('/')
        return segments[-1]

    name = cached_property(get_name)


class OleStream(OleStorageItem):

    def open(self):
        return self.olefile.openstream(self.path)


class OleStorage(OleStorageItem):
    ''' Create an OleStorage instance.

    :param olefile: an OLE2 Compound Binary File.
    :type olefile: an OleFileIO instance or an argument to OleFileIO()
    :param path: internal path in the olefile. Should not end with '/'.
    :raises: `InvalidOleStorageError` when `olefile` is not valid OLE2 format.
    '''

    def __init__(self, olefile, path='', parent=None):
        if not hasattr(olefile, 'openstream'):
            isOleFile = import_isOleFile()
            OleFileIO = import_OleFileIO()

            if not isOleFile(olefile):
                errormsg = 'Not an OLE2 Compound Binary File.'
                raise InvalidOleStorageError(errormsg)
            olefile = OleFileIO(olefile)
        OleStorageItem.__init__(self, olefile, path, parent)

    def __iter__(self):
        return olefile_listdir(self.olefile, self.path)

    def __getitem__(self, name):
        if self.path == '' or self.path == '/':
            path = name
        else:
            path = self.path + '/' + name
        if not self.olefile.exists(path):
            raise KeyError('%s not found' % path)
        t = self.olefile.get_type(path)
        if t == 1:  # Storage
            return OleStorage(self.olefile, path, self)
        elif t == 2:  # Stream
            return OleStream(self.olefile, path, self)
        else:
            raise KeyError('%s is invalid' % path)

    def close(self):
        # if this is root, close underlying olefile
        if self.path == '':
            # old version of OleFileIO has no close()
            if hasattr(self.olefile, 'close'):
                self.olefile.close()


def olefile_listdir(olefile, path):
    if path == '' or path == '/':
        # we use a list instead of a set
        # for python 2.3 compatibility
        yielded = []

        for stream in olefile.listdir():
            top_item = stream[0]
            if top_item in yielded:
                continue
            yielded.append(top_item)
            yield top_item
        return

    if not olefile.exists(path):
        raise IOError('%s not exists' % path)
    if olefile.get_type(path) != 1:
        raise IOError('%s not a storage' % path)
    path_segments = path.split('/')
    for stream in olefile.listdir():
        if len(stream) == len(path_segments) + 1:
            if stream[:-1] == path_segments:
                yield stream[-1]



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals


class STARTEVENT:
    pass


class ENDEVENT:
    pass


def prefix_event(level_prefixed_items, root_item=None):
    ''' convert iterable of (level, item) into iterable of (event, item)
    '''
    baselevel = None
    stack = [root_item]
    for level, item in level_prefixed_items:
        if baselevel is None:
            baselevel = level
            level = 0
        else:
            level -= baselevel

        while level + 1 < len(stack):
            yield ENDEVENT, stack.pop()
        while len(stack) < level + 1:
            raise Exception('invalid level: %d, %d, %s' %
                            (level, len(stack) - 1, item))
        assert(len(stack) == level + 1)

        stack.append(item)
        yield STARTEVENT, item

    while 1 < len(stack):
        yield ENDEVENT, stack.pop()


def prefix_ancestors(event_prefixed_items, root_item=None):
    ''' convert iterable of (event, item) into iterable of (ancestors, item)
    '''
    stack = [root_item]
    for event, item in event_prefixed_items:
        if event is STARTEVENT:
            yield stack, item
            stack.append(item)
        elif event is ENDEVENT:
            stack.pop()


def prefix_ancestors_from_level(level_prefixed_items, root_item=None):
    ''' convert iterable of (level, item) into iterable of (ancestors, item)

        @param level_prefixed items: iterable of tuple(level, item)
        @return iterable of tuple(ancestors, item)
    '''
    baselevel = None
    stack = [root_item]
    for level, item in level_prefixed_items:
        if baselevel is None:
            baselevel = level
            level = 0
        else:
            level -= baselevel

        while level + 1 < len(stack):
            stack.pop()
        while len(stack) < level + 1:
            raise Exception('invalid level: %d, %d, %s' %
                            (level, len(stack) - 1, item))
        assert(len(stack) == level + 1)

        yield stack, item
        stack.append(item)


def build_subtree(event_prefixed_items):
    ''' build a tree from (event, item) stream

        Example Scenario::

           ...
           (STARTEVENT, rootitem)          # should be consumed by the caller
           --- call build_subtree() ---
           (STARTEVENT, child1)            # consumed by build_subtree()
           (STARTEVENT, grandchild)        # (same)
           (ENDEVENT, grandchild)          # (same)
           (ENDEVENT, child1)              # (same)
           (STARTEVENT, child2)            # (same)
           (ENDEVENT, child2)              # (same)
           (ENDEVENT, rootitem)            # same, buildsubtree() returns
           --- build_subtree() returns ---
           (STARTEVENT, another_root)
           ...

        result will be (rootitem, [(child1, [(grandchild, [])]),
                                   (child2, [])])

    '''
    childs = []
    for event, item in event_prefixed_items:
        if event == STARTEVENT:
            childs.append(build_subtree(event_prefixed_items))
        elif event == ENDEVENT:
            return item, childs


def iter_subevents(event_prefixed_items):
    level = 0
    for event, item in event_prefixed_items:
        yield event, item
        if event is STARTEVENT:
            level += 1
        elif event is ENDEVENT:
            if level > 0:
                level -= 1
            else:
                return


def tree_events(rootitem, childs):
    ''' generate tuples of (event, item) from a tree
    '''
    yield STARTEVENT, rootitem
    for k in tree_events_multi(childs):
        yield k
    yield ENDEVENT, rootitem


def tree_events_multi(trees):
    ''' generate tuples of (event, item) from trees
    '''
    for rootitem, childs in trees:
        for k in tree_events(rootitem, childs):
            yield k



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import contextmanager
from functools import partial
import codecs
import logging
import os
import shlex
import shutil
import subprocess
import sys
import tempfile

from .importhelper import pkg_resources_filename


PY3 = sys.version_info.major == 3
logger = logging.getLogger(__name__)


class NIL:
    pass


class cached_property(object):

    def __init__(self, func):
        self.func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, NIL)
        if value is NIL:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value


def generate_json_array(tokens):
    ''' generate json array with given tokens '''
    first = True
    for token in tokens:
        if first:
            yield '[\n'
            first = False
        else:
            yield ',\n'
        yield token
    yield '\n]'


class JsonObjects(object):

    def __init__(self, objects, object_to_json):
        self.objects = objects
        self.object_to_json = object_to_json

    def generate(self, **kwargs):
        kwargs.setdefault('sort_keys', True)
        kwargs.setdefault('indent', 2)

        tokens = (self.object_to_json(obj, **kwargs)
                  for obj in self.objects)
        return generate_json_array(tokens)

    def open(self, **kwargs):
        chunks = self.generate(**kwargs)
        chunks = (chunk.encode('utf-8') for chunk in chunks)
        return GeneratorReader(chunks)

    def dump(self, outfile, **kwargs):
        for s in self.generate(**kwargs):
            outfile.write(s)


def unicode_escape(s):
    '''
    Escape a string.

    :param s:
        a string to escape
    :type s:
        unicode
    :returns:
        escaped string
    :rtype:
        unicode
    '''
    return s.encode('unicode_escape').decode('utf-8')


def unicode_unescape(s):
    '''
    Unescape a string.

    :param s:
        a string to unescape
    :type s:
        unicode
    :returns:
        unescaped string
    :rtype:
        unicode
    '''
    return s.encode('utf-8').decode('unicode_escape')


def transcode(backend_stream, backend_encoding, frontend_encoding,
              errors='strict'):
    enc = codecs.getencoder(frontend_encoding)
    dec = codecs.getdecoder(frontend_encoding)
    rd = codecs.getreader(backend_encoding)
    wr = codecs.getwriter(backend_encoding)
    return codecs.StreamRecoder(backend_stream, enc, dec, rd, wr, errors)


def transcoder(backend_encoding, frontend_encoding, errors='strict'):
    return partial(transcode,
                   backend_encoding=backend_encoding,
                   frontend_encoding=frontend_encoding,
                   errors=errors)


class GeneratorReader(object):
    ''' convert a string generator into file-like reader

        def gen():
            yield b'hello'
            yield b'world'

        f = GeneratorReader(gen())
        assert 'hell' == f.read(4)
        assert 'oworld' == f.read()
    '''

    def __init__(self, gen):
        self.gen = gen
        self.buffer = b''

    def read(self, size=None):
        if size is None:
            d, self.buffer = self.buffer, b''
            return d + b''.join(self.gen)

        for data in self.gen:
            self.buffer += data
            bufsize = len(self.buffer)
            if bufsize >= size:
                size = min(bufsize, size)
                d, self.buffer = self.buffer[:size], self.buffer[size:]
                return d

        d, self.buffer = self.buffer, b''
        return d

    def close(self):
        self.gen = self.buffer = None


class GeneratorTextReader(object):
    ''' convert a string generator into file-like reader

        def gen():
            yield 'hello'
            yield 'world'

        f = GeneratorTextReader(gen())
        assert 'hell' == f.read(4)
        assert 'oworld' == f.read()
    '''

    def __init__(self, gen):
        self.gen = gen
        self.buffer = ''

    def read(self, size=None):
        if size is None:
            d = self.buffer
            self.buffer = ''
            return d + ''.join(self.gen)

        for data in self.gen:
            self.buffer += data
            bufsize = len(self.buffer)
            if bufsize >= size:
                size = min(bufsize, size)
                d, self.buffer = self.buffer[:size], self.buffer[size:]
                return d

        d = self.buffer
        self.buffer = ''
        return d

    def close(self):
        self.gen = self.buffer = None


@contextmanager
def hwp5_resources_path(res_path):
    try:
        path = pkg_resources_filename('hwp5', res_path)
    except Exception:
        logger.info('%s: pkg_resources_filename failed; using resource_stream',
                    res_path)
        with mkstemp_open() as (path, g):
            import pkg_resources
            f = pkg_resources.resource_stream('hwp5', res_path)
            try:
                shutil.copyfileobj(f, g)
                g.close()
                yield path
            finally:
                f.close()
    else:
        yield path


def make_open_dest_file(path):
    if path:
        @contextmanager
        def open_dest_path():
            with open(path, 'wb') as f:
                yield f
        return open_dest_path
    else:
        if PY3:
            @contextmanager
            def open_stdout():
                yield sys.stdout.buffer
            return open_stdout
        else:
            @contextmanager
            def open_stdout():
                yield sys.stdout
            return open_stdout


def wrap_open_dest_for_tty(open_dest, wrappers):
    @contextmanager
    def open_dest_wrapped():
        with open_dest() as output:
            if output.isatty():
                with cascade_contextmanager_filters(output,
                                                    wrappers) as output:
                    yield output
            else:
                yield output
    return open_dest_wrapped


def wrap_open_dest(open_dest, wrappers):
    @contextmanager
    def open_dest_wrapped():
        with open_dest() as output:
            with cascade_contextmanager_filters(output, wrappers) as output:
                yield output
    return open_dest_wrapped


@contextmanager
def cascade_contextmanager_filters(arg, filters):
    if len(filters) == 0:
        yield arg
    else:
        flt, filters = filters[0], filters[1:]
        with flt(arg) as ret:
            with cascade_contextmanager_filters(ret, filters) as ret:
                yield ret


@contextmanager
def null_contextmanager_filter(output):
    yield output


def output_thru_subprocess(cmd):
    @contextmanager
    def filter(output):
        logger.debug('%r', cmd)
        try:
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=output)
        except Exception as e:
            logger.error('%r: %s', ' '.join(cmd), e)
            yield output
        else:
            try:
                yield p.stdin
            except IOError as e:
                import errno
                if e.errno != errno.EPIPE:
                    raise
            finally:
                p.stdin.close()
                p.wait()
                retcode = p.returncode
                logger.debug('%r exit %d', cmd, retcode)
    return filter


def xmllint(c14n=False, encode=None, format=False, nonet=True):
    cmd = ['xmllint']
    if c14n:
        cmd.append('--c14n')
    if encode:
        cmd += ['--encode', encode]
    if format:
        cmd.append('--format')
    if nonet:
        cmd.append('--nonet')
    cmd.append('-')
    return output_thru_subprocess(cmd)


def syntaxhighlight(mimetype):
    try:
        return syntaxhighlight_pygments(mimetype)
    except Exception as e:
        logger.info(e)
        return null_contextmanager_filter


def syntaxhighlight_pygments(mimetype):
    from pygments import highlight
    from pygments.lexers import get_lexer_for_mimetype
    from pygments.formatters import TerminalFormatter

    lexer = get_lexer_for_mimetype(mimetype, encoding='utf-8')
    formatter = TerminalFormatter(encoding='utf-8')

    @contextmanager
    def filter(output):
        with make_temp_file() as f:
            yield f
            f.seek(0)
            code = f.read()
        highlight(code, lexer, formatter, output)
    return filter


@contextmanager
def make_temp_file():
    fd, name = tempfile.mkstemp()
    with unlink_path(name):
        with os.fdopen(fd, 'w+') as f:
            yield f


@contextmanager
def unlink_path(path):
    import os
    try:
        yield
    finally:
        os.unlink(path)


def pager():
    pager_cmd = os.environ.get('PAGER')
    if pager_cmd:
        pager_cmd = shlex.split(pager_cmd)
        return output_thru_subprocess(pager_cmd)
    return pager_less


pager_less = output_thru_subprocess(['less', '-R'])


@contextmanager
def mkstemp_open(*args, **kwargs):

    if (kwargs.get('text', False) or (len(args) >= 4 and args[3])):
        text = True
    else:
        text = False

    mode = 'w+' if text else 'wb+'
    fd, path = tempfile.mkstemp(*args, **kwargs)
    try:
        f = os.fdopen(fd, mode)
        try:
            yield path, f
        finally:
            try:
                f.close()
            except Exception:
                pass
    finally:
        unlink_or_warning(path)


def unlink_or_warning(path):
    try:
        os.unlink(path)
    except Exception as e:
        logger.exception(e)
        logger.warning('%s cannot be deleted', path)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''HWPv5 to HTML converter

Usage::

    hwp5html [options] <hwp5file>
    hwp5html [options] <hwp5file> --html
    hwp5html [options] <hwp5file> --css
    hwp5html -h | --help
    hwp5html --version

Options::

    -h --help           Show this screen
    --version           Show version
    --loglevel=<level>  Set log level.
    --logfile=<file>    Set log file.

    --output=<output>   Output file / directory
'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import contextmanager
from contextlib import closing
from functools import partial
import gettext
import io
import logging
import os.path
import shutil
import sys

from .transforms import BaseTransform
from .utils import cached_property


PY3 = sys.version_info.major == 3
logger = logging.getLogger(__name__)
locale_dir = os.path.join(os.path.dirname(__file__), 'locale')
locale_dir = os.path.abspath(locale_dir)
t = gettext.translation('hwp5html', locale_dir, fallback=True)
if PY3:
    _ = t.gettext
else:
    _ = t.ugettext


RESOURCE_PATH_XSL_CSS = 'xsl/hwp5css.xsl'
RESOURCE_PATH_XSL_XHTML = 'xsl/hwp5html.xsl'


class HTMLTransform(BaseTransform):

    @property
    def transform_hwp5_to_css(self):
        '''
        >>> T.transform_hwp5_to_css(hwp5file, 'styles.css')
        '''
        transform_xhwp5 = self.transform_xhwp5_to_css
        return self.make_transform_hwp5(transform_xhwp5)

    @property
    def transform_hwp5_to_xhtml(self):
        '''
        >>> T.transform_hwp5_to_xhtml(hwp5file, 'index.xhtml')
        '''
        transform_xhwp5 = self.transform_xhwp5_to_xhtml
        return self.make_transform_hwp5(transform_xhwp5)

    def transform_hwp5_to_dir(self, hwp5file, outdir):
        '''
        >>> T.transform_hwp5_to_dir(hwp5file, 'output')
        '''
        with self.transformed_xhwp5_at_temp(hwp5file) as xhwp5path:
            self.transform_xhwp5_to_dir(xhwp5path, outdir)

        bindata_dir = os.path.join(outdir, 'bindata')
        self.extract_bindata_dir(hwp5file, bindata_dir)

    @cached_property
    def transform_xhwp5_to_css(self):
        '''
        >>> T.transform_xhwp5_to_css('hwp5.xml', 'styles.css')
        '''
        resource_path = RESOURCE_PATH_XSL_CSS
        return self.make_xsl_transform(resource_path)

    @cached_property
    def transform_xhwp5_to_xhtml(self):
        '''
        >>> T.transform_xhwp5_to_xhtml('hwp5.xml', 'index.xhtml')
        '''
        resource_path = RESOURCE_PATH_XSL_XHTML
        return self.make_xsl_transform(resource_path)

    def transform_xhwp5_to_dir(self, xhwp5path, outdir):
        '''
        >>> T.transform_xhwp5_to_dir('hwp5.xml', 'output')
        '''
        html_path = os.path.join(outdir, 'index.xhtml')
        with io.open(html_path, 'wb') as f:
            self.transform_xhwp5_to_xhtml(xhwp5path, f)

        css_path = os.path.join(outdir, 'styles.css')
        with io.open(css_path, 'wb') as f:
            self.transform_xhwp5_to_css(xhwp5path, f)

    def extract_bindata_dir(self, hwp5file, bindata_dir):
        if 'BinData' not in hwp5file:
            return
        bindata_stg = hwp5file['BinData']
        if not os.path.exists(bindata_dir):
            os.mkdir(bindata_dir)

        from hwp5.storage import unpack
        unpack(bindata_stg, bindata_dir)


def main():
    from docopt import docopt

    from . import __version__ as version
    from .dataio import ParseError
    from .errors import InvalidHwp5FileError
    from .proc import rest_to_docopt
    from .proc import init_logger
    from .utils import make_open_dest_file
    from .xmlmodel import Hwp5File

    doc = rest_to_docopt(__doc__)
    args = docopt(doc, version=version)
    init_logger(args)

    hwp5path = args['<hwp5file>']

    html_transform = HTMLTransform()

    open_dest = make_open_dest_file(args['--output'])
    if args['--css']:
        transform = html_transform.transform_hwp5_to_css
        open_dest = wrap_for_css(open_dest)
    elif args['--html']:
        transform = html_transform.transform_hwp5_to_xhtml
        open_dest = wrap_for_xml(open_dest)
    else:
        transform = html_transform.transform_hwp5_to_dir
        dest_path = args['--output']
        if not dest_path:
            dest_path = os.path.splitext(os.path.basename(hwp5path))[0]
        open_dest = partial(open_dir, dest_path)

    try:
        with closing(Hwp5File(hwp5path)) as hwp5file:
            with open_dest() as dest:
                transform(hwp5file, dest)
    except ParseError as e:
        e.print_to_logger(logger)
    except InvalidHwp5FileError as e:
        logger.error('%s', e)
        sys.exit(1)


@contextmanager
def open_dir(path):
    if os.path.exists(path):
        shutil.rmtree(path)
    os.mkdir(path)
    yield path


def wrap_for_css(open_dest):
    from .utils import wrap_open_dest_for_tty
    from .utils import pager
    from .utils import syntaxhighlight
    return wrap_open_dest_for_tty(open_dest, [
        pager(),
        syntaxhighlight('text/css'),
    ])


def wrap_for_xml(open_dest):
    from .utils import wrap_open_dest_for_tty
    from .utils import pager
    from .utils import syntaxhighlight
    from .utils import xmllint
    return wrap_open_dest_for_tty(open_dest, [
        pager(),
        syntaxhighlight('application/xml'),
        xmllint(format=True, nonet=True),
    ])



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from collections import deque
from itertools import chain
from pprint import pformat
from tempfile import TemporaryFile
import base64
import logging
import sys

from . import binmodel
from . import filestructure
from .binmodel.controls import SectionDef
from .binmodel.controls import TableControl
from .binmodel.controls import GShapeObjectControl
from .binmodel import BinData
from .binmodel import ListHeader
from .binmodel import Paragraph
from .binmodel import Text
from .binmodel import ShapeComponent
from .binmodel import TableBody
from .binmodel import TableCell
from .binmodel import ParaText
from .binmodel import ParaLineSeg
from .binmodel import ParaCharShape
from .binmodel import LineSeg
from .binmodel import ParaRangeTag
from .binmodel import Field
from .binmodel import ControlChar
from .binmodel import Control
from .charsets import tokenize_unicode_by_lang
from .dataio import Struct
from .filestructure import VERSION
from .treeop import STARTEVENT, ENDEVENT
from .treeop import prefix_event
from .treeop import build_subtree
from .treeop import tree_events
from .treeop import tree_events_multi
from .xmlformat import startelement
from .xmlformat import xmlevents_to_bytechunks


PY3 = sys.version_info.major == 3
if PY3:
    basestring = str
    unichr = chr


logger = logging.getLogger(__name__)


def give_elements_unique_id(event_prefixed_mac):
    paragraph_id = 0
    table_id = 0
    gshape_id = 0
    shape_id = 0
    for event, item in event_prefixed_mac:
        (model, attributes, context) = item
        if event == STARTEVENT:
            if model == Paragraph:
                attributes['paragraph_id'] = paragraph_id
                paragraph_id += 1
            elif model == TableControl:
                attributes['table_id'] = table_id
                table_id += 1
            elif model == GShapeObjectControl:
                attributes['gshape_id'] = gshape_id
                gshape_id += 1
            elif model == ShapeComponent:
                attributes['shape_id'] = shape_id
                shape_id += 1
        yield event, item


def make_ranged_shapes(shapes):
    last = None
    for item in shapes:
        if last is not None:
            yield (last[0], item[0]), last[1]
        last = item
    yield (item[0], 0x7fffffff), item[1]


def split_and_shape(chunks, ranged_shapes):
    try:
        (chunk_start, chunk_end), chunk_attr, chunk = next(chunks)
    except StopIteration:
        return
    for (shape_start, shape_end), shape in ranged_shapes:
        while True:
            # case 0: chunk has left intersection
            #        vvvv
            #      ----...
            if chunk_start < shape_start:
                assert False

            # case 1: chunk is far right: get next shape
            #         vvvv
            #             ----
            if shape_end <= chunk_start:        # (1)
                break

            assert chunk_start < shape_end      # by (1)
            assert shape_start <= chunk_start
            # case 2: chunk has left intersection
            #         vvvv
            #         ..----
            if shape_end < chunk_end:           # (2)
                prev = ((chunk_start, shape_end),
                        chunk[:shape_end - chunk_start])
                nexT = ((shape_end, chunk_end),
                        chunk[shape_end - chunk_start:])
                (chunk_start, chunk_end), chunk = prev
            else:
                nexT = None

            assert chunk_end <= shape_end       # by (2)
            yield (chunk_start, chunk_end), (shape, chunk_attr), chunk

            if nexT is not None:
                (chunk_start, chunk_end), chunk = nexT
                continue

            try:
                (chunk_start, chunk_end), chunk_attr, chunk = next(chunks)
            except StopIteration:
                return


def line_segmented(chunks, ranged_linesegs):
    prev_lineseg = None
    line = None
    for ((chunk_start, chunk_end),
         (lineseg, chunk_attr),
         chunk) in split_and_shape(chunks, ranged_linesegs):
        if lineseg is not prev_lineseg:
            if line is not None:
                yield prev_lineseg, line
            line = []
        line.append(((chunk_start, chunk_end), chunk_attr, chunk))
        prev_lineseg = lineseg
    if line is not None:
        yield prev_lineseg, line


def make_texts_linesegmented_and_charshaped(event_prefixed_mac):
    ''' lineseg/charshaped text chunks '''

    stack = []  # stack of ancestor Paragraphs
    for event, item in event_prefixed_mac:
        model, attributes, context = item
        if model is Paragraph:
            if event == STARTEVENT:
                stack.append(dict())
                yield STARTEVENT, item
            else:
                paratext = stack[-1].get(ParaText)
                paracharshape = stack[-1].get(ParaCharShape)
                paralineseg = stack[-1].get(ParaLineSeg)
                # TODO: RangeTags are not used for now
                # pararangetag = stack[-1].get(ParaRangeTag)
                if paratext is None:
                    paratext = (ParaText,
                                dict(chunks=[((0, 0), '')]),
                                dict(context))
                for x in merge_paragraph_text_charshape_lineseg(paratext,
                                                                paracharshape,
                                                                paralineseg):
                    yield x

                yield ENDEVENT, (model, attributes, context)
                stack.pop()
        elif model in (ParaText, ParaCharShape, ParaLineSeg, ParaRangeTag):
            if event == STARTEVENT:
                stack[-1][model] = model, attributes, context
        else:
            yield event, (model, attributes, context)


def merge_paragraph_text_charshape_lineseg(paratext, paracharshape,
                                           paralineseg):

    paratext_model, paratext_attributes, paratext_context = paratext

    chunks = ((range, None, chunk)
              for range, chunk in paratext_attributes['chunks'])
    charshapes = paracharshape[1]['charshapes']
    shaped_chunks = split_and_shape(chunks, make_ranged_shapes(charshapes))

    if paralineseg:
        paralineseg_content = paralineseg[1]
        paralineseg_context = paralineseg[2]
    else:
        # 배포용 문서의 더미 BodyText 에는 LineSeg 정보가 없음
        # (see https://github.com/mete0r/pyhwp/issues/33)
        # 더미 LineSeg를 만들어 준다
        lineseg = dict(chpos=0, y=0, height=0, height2=0, height85=0,
                       space_below=0, x=0, width=0, a8=0, flags=0)
        paralineseg_content = dict(linesegs=[lineseg])
        paralineseg_context = dict()
    linesegs = ((lineseg['chpos'], lineseg)
                for lineseg in paralineseg_content['linesegs'])
    lined_shaped_chunks = line_segmented(shaped_chunks,
                                         make_ranged_shapes(linesegs))
    for lineseg_content, shaped_chunks in lined_shaped_chunks:
        lineseg = (LineSeg, lineseg_content, paralineseg_context)
        chunk_events = range_shaped_textchunk_events(paratext_context,
                                                     shaped_chunks)
        for x in wrap_modelevents(lineseg, chunk_events):
            yield x


def range_shaped_textchunk_events(paratext_context, range_shaped_textchunks):
    for (startpos, endpos), (shape, none), chunk in range_shaped_textchunks:
        if isinstance(chunk, basestring):
            textitem = (Text,
                        dict(text=chunk, charshape_id=shape),
                        paratext_context)
            yield STARTEVENT, textitem
            yield ENDEVENT, textitem
        elif isinstance(chunk, dict):
            code = chunk['code']
            uch = unichr(code)
            name = ControlChar.get_name_by_code(code)
            kind = ControlChar.kinds[uch]
            chunk_attributes = dict(name=name,
                                    code=code,
                                    kind=kind,
                                    charshape_id=shape)
            if code in (0x9, 0xa, 0xd):  # http://www.w3.org/TR/xml/#NT-Char
                chunk_attributes['char'] = uch
            ctrlch = (ControlChar, chunk_attributes, paratext_context)
            yield STARTEVENT, ctrlch
            yield ENDEVENT, ctrlch


def wrap_section(event_prefixed_mac, sect_id=None):
    ''' wrap a section with SectionDef '''
    starting_buffer = list()
    started = False
    sectiondef = None
    for event, item in event_prefixed_mac:
        if started:
            yield event, item
        else:
            model, attributes, context = item
            if model is SectionDef and event is STARTEVENT:
                sectiondef, sectdef_child = build_subtree(event_prefixed_mac)
                if sect_id is not None:
                    attributes['section_id'] = sect_id
                yield STARTEVENT, sectiondef
                for k in tree_events_multi(sectdef_child):
                    yield k
                for evented_item in starting_buffer:
                    yield evented_item
                started = True
            else:
                starting_buffer.append((event, item))
    yield ENDEVENT, sectiondef


class ColumnSet:
    pass


def wrap_columns(event_prefixed_mac):

    stack = []

    for event, item in event_prefixed_mac:
        model, attributes, context = item

        if model is Paragraph:
            if event is STARTEVENT:

                split = attributes['split']
                split = Paragraph.SplitFlags(split)

                if split.new_columnsdef:
                    if stack[-1][0] is ColumnSet:
                        yield ENDEVENT, stack.pop()

                    columns = (ColumnSet, {}, {})
                    stack.append(columns)
                    yield STARTEVENT, columns

        else:
            if event is STARTEVENT:
                stack.append(item)
            else:
                if model != stack[-1][0]:
                    assert stack[-1][0] is ColumnSet
                    yield ENDEVENT, stack.pop()
                stack.pop()

        yield event, item


def make_extended_controls_inline(event_prefixed_mac, stack=None):
    ''' inline extended-controls into paragraph texts '''
    if stack is None:
        stack = []  # stack of ancestor Paragraphs
    for event, item in event_prefixed_mac:
        model, attributes, context = item
        if model is Paragraph:
            for x in meci_paragraph(event, stack, item):
                yield x
        elif model is ControlChar:
            for x in meci_controlchar(event, stack, item, attributes):
                yield x
        elif issubclass(model, Control) and event == STARTEVENT:
            control_subtree = build_subtree(event_prefixed_mac)
            paragraph = stack[-1]
            paragraph_controls = paragraph.setdefault(Control, [])
            paragraph_controls.append(control_subtree)
        else:
            yield event, item


def meci_paragraph(event, stack, item):
    if event == STARTEVENT:
        stack.append(dict())
        yield STARTEVENT, item
    else:
        yield ENDEVENT, item
        stack.pop()


def meci_controlchar(event, stack, item, attributes):
    if event is STARTEVENT:
        if attributes['kind'] is ControlChar.EXTENDED:
            paragraph = stack[-1]
            paragraph_controls = paragraph.get(Control)
            control_subtree = paragraph_controls.pop(0)
            tev = tree_events(*control_subtree)
            # to evade the Control/STARTEVENT trigger
            # in parse_models_pass3()
            yield next(tev)

            for k in make_extended_controls_inline(tev, stack):
                yield k
        else:
            yield STARTEVENT, item
            yield ENDEVENT, item


def make_paragraphs_children_of_listheader(event_prefixed_mac,
                                           parentmodel=ListHeader,
                                           childmodel=Paragraph):
    ''' make paragraphs children of the listheader '''
    stack = []
    level = 0
    for event, item in event_prefixed_mac:
        model, attributes, context = item
        if event is STARTEVENT:
            level += 1
        if len(stack) > 0 and ((event is STARTEVENT
                                and stack[-1][0] == level
                                and model is not childmodel) or
                               (event is ENDEVENT
                                and stack[-1][0] - 1 == level)):
            lh_level, lh_item = stack.pop()
            yield ENDEVENT, lh_item

        if issubclass(model, parentmodel):
            if event is STARTEVENT:
                stack.append((level, item))
                yield event, item
            else:
                pass
        else:
            yield event, item

        if event is ENDEVENT:
            level -= 1


def match_field_start_end(event_prefixed_mac):
    stack = []
    for event, item in event_prefixed_mac:
        (model, attributes, context) = item
        if issubclass(model, Field):
            for x in mfse_field(event, stack, item):
                yield x
        elif model is LineSeg:
            for x in mfse_lineseg(event, stack, item):
                yield x
        elif model is ControlChar and attributes['name'] == 'FIELD_END':
            for x in mfse_field_end(event, stack, item):
                yield x
        else:
            yield event, item


def mfse_field(event, stack, item):
    if event is STARTEVENT:
        stack.append(item)
        yield event, item
    else:
        pass


def mfse_lineseg(event, stack, item):
    if event is ENDEVENT:
        # fields still not closed; temporarily close them
        for field_item in reversed(stack):
            yield ENDEVENT, field_item
        yield event, item
    elif event is STARTEVENT:
        yield event, item
        # fields temporarily closed; open them again
        for field_item in stack:
            yield STARTEVENT, field_item


def mfse_field_end(event, stack, item):
    if event is ENDEVENT:
        if len(stack) > 0:
            yield event, stack.pop()
        else:
            logger.warning('unmatched field end')


class TableRow:
    pass


ROW_OPEN = 1
ROW_CLOSE = 2


def restructure_tablebody(event_prefixed_mac):
    ''' Group table columns in each rows and wrap them with TableRow. '''
    stack = []
    for event, item in event_prefixed_mac:
        (model, attributes, context) = item
        if model is TableBody:
            for x in rstbody_tablebody(event, stack, item, attributes,
                                       context):
                yield x
        elif model is TableCell:
            for x in rstbody_tablecell(event, stack, item):
                yield x
        else:
            yield event, item


def rstbody_tablebody(event, stack, item, attributes, context):
    if event is STARTEVENT:
        rowcols = deque()
        for cols in attributes.pop('rowcols'):
            if cols == 1:
                rowcols.append(ROW_OPEN | ROW_CLOSE)
            else:
                rowcols.append(ROW_OPEN)
                for i in range(0, cols - 2):
                    rowcols.append(0)
                rowcols.append(ROW_CLOSE)
        stack.append((context, rowcols))
        yield event, item
    else:
        yield event, item
        stack.pop()


def rstbody_tablecell(event, stack, item):
    table_context, rowcols = stack[-1]
    row_context = dict(table_context)
    if event is STARTEVENT:
        how = rowcols[0]
        if how & ROW_OPEN:
            yield STARTEVENT, (TableRow, dict(), row_context)
    yield event, item
    if event is ENDEVENT:
        how = rowcols.popleft()
        if how & ROW_CLOSE:
            yield ENDEVENT, (TableRow, dict(), row_context)


def tokenize_text_by_lang(event_prefixed_mac):
    ''' Group table columns in each rows and wrap them with TableRow. '''
    for event, item in event_prefixed_mac:
        (model, attributes, context) = item
        if model is Text:
            if event is STARTEVENT:
                charshape_id = attributes['charshape_id']
                for lang, text in tokenize_unicode_by_lang(attributes['text']):
                    token = (Text, {
                        'charshape_id': charshape_id,
                        'lang': lang,
                        'text': text,
                    }, context)
                    yield STARTEVENT, token
                    yield ENDEVENT, token
        else:
            yield event, item


def embed_bindata(event_prefixed_mac, bindata):
    for event, item in event_prefixed_mac:
        (model, attributes, context) = item
        if event is STARTEVENT and model is BinData:
            if attributes['flags'].storage is BinData.StorageType.EMBEDDING:
                name = ('BIN%04X' % attributes['bindata']['storage_id']
                        + '.'
                        + attributes['bindata']['ext'])
                bin_stream = bindata[name].open()
                try:
                    binary = bin_stream.read()
                finally:
                    bin_stream.close()
                b64 = base64.b64encode(binary)
                b64 = b64.decode('ascii')
                truncated = []
                while b64:
                    if len(b64) > 64:
                        truncated.append(b64[:64])
                        b64 = b64[64:]
                    else:
                        truncated.append(b64)
                        b64 = ''
                b64 = '\n'.join(truncated)
                b64 = '\n' + b64 + '\n'
                attributes['bindata']['<text>'] = b64
                attributes['bindata']['inline'] = 'true'
        yield event, item


def prefix_binmodels_with_event(context, models):
    level_prefixed = ((model['level'],
                       (model['type'], model['content'], context))
                      for model in models)
    return prefix_event(level_prefixed)


def wrap_modelevents(wrapper_model, modelevents):
    yield STARTEVENT, wrapper_model
    for mev in modelevents:
        yield mev
    yield ENDEVENT, wrapper_model


def modelevents_to_xmlevents(modelevents):
    for event, (model, attributes, context) in modelevents:
        try:
            if event is STARTEVENT:
                for x in startelement(context, (model, attributes)):
                    yield x
            elif event is ENDEVENT:
                yield ENDEVENT, model.__name__
        except:
            logger.error('model: %s', pformat({
                'event': event,
                'model': model,
                'attributes': attributes,
                'context': context
            }))
            raise


class XmlEvents(object):

    def __init__(self, events):
        self.events = events

    def __iter__(self):
        return modelevents_to_xmlevents(self.events)

    def bytechunks(self, xml_declaration=True, **kwargs):
        encoding = kwargs.get('xml_encoding', 'utf-8')
        if xml_declaration:
            yield '<?xml version="1.0" encoding="{}"?>\n'.format(
                encoding
            ).encode(
                encoding
            )
        bytechunks = xmlevents_to_bytechunks(self, encoding)
        for chunk in bytechunks:
            yield chunk

    def dump(self, outfile, **kwargs):
        bytechunks = self.bytechunks(**kwargs)
        for chunk in bytechunks:
            outfile.write(chunk)
        if hasattr(outfile, 'flush'):
            outfile.flush()

    def open(self, **kwargs):
        tmpfile = TemporaryFile()
        try:
            self.dump(tmpfile, **kwargs)
        except:
            tmpfile.close()
            raise

        tmpfile.seek(0)
        return tmpfile


class XmlEventsMixin(object):

    def xmlevents(self, **kwargs):
        return XmlEvents(self.events(**kwargs))


class ModelEventStream(binmodel.ModelStream, XmlEventsMixin):

    def modelevents(self, **kwargs):
        models = self.models(**kwargs)

        # prepare modelevents context
        kwargs.setdefault('version', self.version)
        return prefix_binmodels_with_event(kwargs, models)

    def other_formats(self):
        d = super(ModelEventStream, self).other_formats()
        d['.xml'] = self.xmlevents().open
        return d


class HwpSummaryInfo(filestructure.HwpSummaryInfo, XmlEventsMixin):

    def events(self, **context):
        generator = PropertySetStreamModelEventsGenerator(context)
        events = generator.generateModelEvents(self.propertySetStream)
        element = HwpSummaryInfo, {}, context
        return wrap_modelevents(element, events)


class PropertySetStreamModelEventsGenerator(object):

    def __init__(self, context):
        self.context = context

    def generateModelEvents(self, stream):
        return self.getPropertySetStreamEvents(stream)

    def getPropertySetStreamEvents(self, stream):
        from .msoleprops import PropertySetStream
        sectionEvents = [
            self.getPropertySetEvents(propertyset)
            for propertyset in stream.propertysets
        ]
        events = chain(*sectionEvents)

        content = dict(
            byte_order='{:04x}'.format(
                stream.byteOrder,
            ),
            version=str(stream.version),
            system_identifier='{:08x}'.format(
                stream.systemIdentifier,
            ),
            clsid=str(stream.clsid)
        )
        element = PropertySetStream, content, self.context
        return wrap_modelevents(element, events)

    def getPropertySetEvents(self, propertyset):
        from .msoleprops import PropertySet
        propertyEvents = [
            self.getPropertyEvents(property)
            for property in sorted(
                propertyset.properties,
                key=lambda property: property.desc.offset
            )
        ]
        events = chain(*propertyEvents)

        content = dict(
            fmtid=propertyset.fmtid,
            offset=propertyset.desc.offset,
        )
        element = PropertySet, content, self.context
        return wrap_modelevents(element, events)

    def getPropertyEvents(self, property):
        from .msoleprops import PID_DICTIONARY
        from .msoleprops import Property
        content = dict(
            id=property.desc.id,
            offset=property.desc.offset,
        )
        if property.idLabel is not None:
            content['id_label'] = property.idLabel
        if property.type is not None:
            content['type'] = str(property.type.vt_type.__name__)
            content['type_code'] = '0x{:04x}'.format(property.type.code)
        if property.id == PID_DICTIONARY.id:
            events = self.getDictionaryEvents(property.value)
        else:
            events = ()
            content['value'] = property.value
        element = Property, content, self.context
        return wrap_modelevents(element, events)

    def getDictionaryEvents(self, dictionary):
        events = list(self.getDictionaryEntryEvents(entry)
                      for entry in dictionary.entries)
        return chain(*events)

    def getDictionaryEntryEvents(self, entry):
        from .msoleprops import DictionaryEntry
        content = dict(
            id=entry.id,
            name=entry.name,
        )
        element = DictionaryEntry, content, self.context
        return wrap_modelevents(element, ())


class DocInfo(ModelEventStream):

    def events(self, **kwargs):
        docinfo = DocInfo, dict(), dict()
        events = self.modelevents(**kwargs)
        if 'embedbin' in kwargs:
            events = embed_bindata(events, kwargs['embedbin'])
        events = wrap_modelevents(docinfo, events)
        return events


class Section(ModelEventStream):

    def events(self, **kwargs):
        events = self.modelevents(**kwargs)

        events = make_texts_linesegmented_and_charshaped(events)
        events = make_extended_controls_inline(events)
        events = match_field_start_end(events)
        events = make_paragraphs_children_of_listheader(events)
        events = make_paragraphs_children_of_listheader(events, TableBody,
                                                        TableCell)
        events = restructure_tablebody(events)
        events = tokenize_text_by_lang(events)

        section_idx = kwargs.get('section_idx')
        events = wrap_section(events, section_idx)
        events = wrap_columns(events)

        return events


class Sections(binmodel.Sections, XmlEventsMixin):

    section_class = Section

    def events(self, **kwargs):
        bodytext_events = []
        for idx in self.section_indexes():
            kwargs['section_idx'] = idx
            section = self.section(idx)
            events = section.events(**kwargs)
            bodytext_events.append(events)

        class BodyText(object):
            pass
        bodytext_events = chain(*bodytext_events)
        bodytext = BodyText, dict(), dict()
        return wrap_modelevents(bodytext, bodytext_events)

    def other_formats(self):
        d = super(Sections, self).other_formats()
        d['.xml'] = self.xmlevents().open
        return d


class HwpDoc(Struct):

    def attributes():
        yield VERSION, 'version'
    attributes = staticmethod(attributes)


class Hwp5File(binmodel.Hwp5File, XmlEventsMixin):

    summaryinfo_class = HwpSummaryInfo
    docinfo_class = DocInfo
    bodytext_class = Sections

    def events(self, **kwargs):
        if 'embedbin' in kwargs and kwargs['embedbin'] and 'BinData' in self:
            kwargs['embedbin'] = self['BinData']
        else:
            kwargs.pop('embedbin', None)

        events = chain(self.summaryinfo.events(**kwargs),
                       self.docinfo.events(**kwargs),
                       self.text.events(**kwargs))

        hwpdoc = HwpDoc, dict(version=self.header.version), dict()
        events = wrap_modelevents(hwpdoc, events)

        # for easy references in styles
        events = give_elements_unique_id(events)

        return events



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import re
import sys

from six import with_metaclass

from hwp5.dataio import PrimitiveType
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import UINT8


PY3 = sys.version_info.major == 3
if PY3:
    unichr = chr


class CHID(with_metaclass(PrimitiveType, str)):

    fixed_size = 4

    # Common controls
    GSO = 'gso '
    TBL = 'tbl '
    LINE = '$lin'
    RECT = '$rec'
    ELLI = '$ell'
    ARC = '$arc'
    POLY = '$pol'
    CURV = '$cur'
    EQED = 'eqed'
    PICT = '$pic'
    OLE = '$ole'
    CONTAINER = '$con'

    # Controls
    SECD = 'secd'
    COLD = 'cold'
    HEADER = 'head'
    FOOTER = 'foot'
    FN = 'fn  '
    EN = 'en  '
    ATNO = 'atno'
    NWNO = 'nwno'
    PGHD = 'pghd'
    PGCT = 'pgct'
    PGNP = 'pgnp'
    IDXM = 'idxm'
    BOKM = 'bokm'
    TCPS = 'tcps'
    TDUT = 'tdut'
    TCMT = 'tcmt'

    # Field starts
    FIELD_UNK = '%unk'
    FIELD_DTE = '%dte'
    FIELD_DDT = '%ddt'
    FIELD_PAT = '%pat'
    FIELD_BMK = '%bmk'
    FIELD_MMG = '%mmg'
    FIELD_XRF = '%xrf'
    FIELD_FMU = '%fmu'
    FIELD_CLK = '%clk'
    FIELD_SMR = '%smr'
    FIELD_USR = '%usr'
    FIELD_HLK = '%hlk'
    FIELD_REVISION_SIGN = '%sig'
    FIELD_REVISION_DELETE = '%%*d'
    FIELD_REVISION_ATTACH = '%%*a'
    FIELD_REVISION_CLIPPING = '%%*C'
    FIELD_REVISION_SAWTOOTH = '%%*S'
    FIELD_REVISION_THINKING = '%%*T'
    FIELD_REVISION_PRAISE = '%%*P'
    FIELD_REVISION_LINE = '%%*L'
    FIELD_REVISION_SIMPLECHANGE = '%%*c'
    FIELD_REVISION_HYPERLINK = '%%*h'
    FIELD_REVISION_LINEATTACH = '%%*A'
    FIELD_REVISION_LINELINK = '%%*i'
    FIELD_REVISION_LINETRANSFER = '%%*t'
    FIELD_REVISION_RIGHTMOVE = '%%*r'
    FIELD_REVISION_LEFTMOVE = '%%*l'
    FIELD_REVISION_TRANSFER = '%%*n'
    FIELD_REVISION_SIMPLEINSERT = '%%*e'
    FIELD_REVISION_SPLIT = '%spl'
    FIELD_REVISION_CHANGE = '%%mr'
    FIELD_MEMO = '%%me'
    FIELD_PRIVATE_INFO_SECURITY = '%cpr'

    def decode(bytes, context=None):
        if PY3:
            return (
                chr(bytes[3]) +
                chr(bytes[2]) +
                chr(bytes[1]) +
                chr(bytes[0])
            )
        else:
            return bytes[3] + bytes[2] + bytes[1] + bytes[0]
    decode = staticmethod(decode)


class ControlChar(object):
    class CHAR(object):
        size = 1

    class INLINE(object):
        size = 8

    class EXTENDED(object):
        size = 8
    chars = {0x00: ('NULL', CHAR),
             0x01: ('CTLCHR01', EXTENDED),
             0x02: ('SECTION_COLUMN_DEF', EXTENDED),
             0x03: ('FIELD_START', EXTENDED),
             0x04: ('FIELD_END', INLINE),
             0x05: ('CTLCHR05', INLINE),
             0x06: ('CTLCHR06', INLINE),
             0x07: ('CTLCHR07', INLINE),
             0x08: ('TITLE_MARK', INLINE),
             0x09: ('TAB', INLINE),
             0x0a: ('LINE_BREAK', CHAR),
             0x0b: ('DRAWING_TABLE_OBJECT', EXTENDED),
             0x0c: ('CTLCHR0C', EXTENDED),
             0x0d: ('PARAGRAPH_BREAK', CHAR),
             0x0e: ('CTLCHR0E', EXTENDED),
             0x0f: ('HIDDEN_EXPLANATION', EXTENDED),
             0x10: ('HEADER_FOOTER', EXTENDED),
             0x11: ('FOOT_END_NOTE', EXTENDED),
             0x12: ('AUTO_NUMBER', EXTENDED),
             0x13: ('CTLCHR13', INLINE),
             0x14: ('CTLCHR14', INLINE),
             0x15: ('PAGE_CTLCHR', EXTENDED),
             0x16: ('BOOKMARK', EXTENDED),
             0x17: ('CTLCHR17', EXTENDED),
             0x18: ('HYPHEN', CHAR),
             0x1e: ('NONBREAK_SPACE', CHAR),
             0x1f: ('FIXWIDTH_SPACE', CHAR)}
    names = dict((unichr(code), name) for code, (name, kind) in chars.items())
    kinds = dict((unichr(code), kind) for code, (name, kind) in chars.items())

    def _populate(cls):
        for ch, name in cls.names.items():
            setattr(cls, name, ch)
    _populate = classmethod(_populate)
    REGEX_CONTROL_CHAR = re.compile(b'[\x00-\x1f]\x00')

    def find(cls, data, start_idx):
        while True:
            m = cls.REGEX_CONTROL_CHAR.search(data, start_idx)
            if m is not None:
                i = m.start()
                if i & 1 == 1:
                    start_idx = i + 1
                    continue
                if PY3:
                    char = unichr(data[i])
                else:
                    char = unichr(ord(data[i]))
                size = cls.kinds[char].size
                return i, i + (size * 2)
            data_len = len(data)
            return data_len, data_len
    find = classmethod(find)

    def decode(cls, bytes):
        code = UINT16.decode(bytes[0:2])
        ch = unichr(code)
        if cls.kinds[ch].size == 8:
            bytes = bytes[2:2 + 12]
            if ch == ControlChar.TAB:
                param = dict(width=UINT32.decode(bytes[0:4]),
                             unknown0=UINT8.decode(bytes[4:5]),
                             unknown1=UINT8.decode(bytes[5:6]),
                             unknown2=bytes[6:])
                return dict(code=code, param=param)
            else:
                chid = CHID.decode(bytes[0:4])
                param = bytes[4:12]
                return dict(code=code, chid=chid, param=param)
        else:
            return dict(code=code)
    decode = classmethod(decode)

    def get_kind_by_code(cls, code):
        ch = unichr(code)
        return cls.kinds[ch]
    get_kind_by_code = classmethod(get_kind_by_code)

    def get_name_by_code(cls, code):
        ch = unichr(code)
        return cls.names.get(ch, 'CTLCHR%02x' % code)
    get_name_by_code = classmethod(get_name_by_code)


ControlChar._populate()



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_PAGE_DEF
from hwp5.dataio import Enum
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import HWPUNIT


class PageDef(RecordModel):
    tagid = HWPTAG_PAGE_DEF
    Orientation = Enum(PORTRAIT=0, LANDSCAPE=1)
    BookBinding = Enum(LEFT=0, RIGHT=1, TOP=2, BOTTOM=3)
    Flags = Flags(UINT32,
                  0, Orientation, 'orientation',
                  1, 2, BookBinding, 'bookbinding')

    def attributes(cls):
        yield HWPUNIT, 'width',
        yield HWPUNIT, 'height',
        yield HWPUNIT, 'left_offset',
        yield HWPUNIT, 'right_offset',
        yield HWPUNIT, 'top_offset',
        yield HWPUNIT, 'bottom_offset',
        yield HWPUNIT, 'header_offset',
        yield HWPUNIT, 'footer_offset',
        yield HWPUNIT, 'bookbinding_offset',
        yield cls.Flags, 'attr',
        # yield UINT32, 'attr',
    attributes = classmethod(attributes)

    def getDimension(self):
        width = HWPUNIT(self.paper_width - self.offsetLeft - self.offsetRight)
        height = HWPUNIT(self.paper_height
                         - (self.offsetTop + self.offsetHeader)
                         - (self.offsetBottom + self.offsetFooter))
        if self.attr.landscape:
            return (height, width)
        else:
            return (width, height)
    dimension = property(getDimension)

    def getHeight(self):
        if self.attr.landscape:
            width = HWPUNIT(self.paper_width - self.offsetLeft -
                            self.offsetRight)
            return width
        else:
            height = HWPUNIT(self.paper_height
                             - (self.offsetTop + self.offsetHeader)
                             - (self.offsetBottom + self.offsetFooter))
            return height
    height = property(getHeight)

    def getWidth(self):
        if self.attr.landscape:
            height = HWPUNIT(self.paper_height
                             - (self.offsetTop + self.offsetHeader)
                             - (self.offsetBottom + self.offsetFooter))
            return height
        else:
            width = HWPUNIT(self.paper_width - self.offsetLeft -
                            self.offsetRight)
            return width
    width = property(getWidth)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_BORDER_FILL
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import BYTE
from hwp5.dataio import Flags
# from hwp5.dataio import N_ARRAY
from hwp5.binmodel._shared import Border
from hwp5.binmodel._shared import FillColorPattern
from hwp5.binmodel._shared import FillGradation
from hwp5.binmodel._shared import FillImage


class BorderFill(RecordModel):
    ''' 4.1.5. 테두리/배경 '''

    tagid = HWPTAG_BORDER_FILL

    # 표 19 테두리/배경 속성
    BorderFlags = Flags(UINT16,
                        0, 'effect_3d',
                        1, 'effect_shadow',
                        2, 4, 'slash',
                        5, 6, 'backslash')

    # 표 23 채우기 정보
    FillFlags = Flags(UINT32,
                      0, 'colorpattern',
                      1, 'image',
                      2, 'gradation')

    def attributes(cls):
        ''' 표 18 테두리/배경 속성 '''
        yield cls.BorderFlags, 'borderflags'
        yield Border, 'left',
        yield Border, 'right',
        yield Border, 'top',
        yield Border, 'bottom',
        yield Border, 'diagonal'
        yield cls.FillFlags, 'fillflags'

        def fill_colorpattern(context, values):
            ''' fillflags.fill_colorpattern '''
            return values['fillflags'].colorpattern

        def fill_image(context, values):
            ''' fillflags.fill_image '''
            return values['fillflags'].image

        def fill_gradation(context, values):
            ''' fillflags.fill_gradation '''
            return values['fillflags'].gradation

        yield dict(type=FillColorPattern, name='fill_colorpattern',
                   condition=fill_colorpattern)
        yield dict(type=FillGradation, name='fill_gradation',
                   condition=fill_gradation)
        yield dict(type=FillImage, name='fill_image',
                   condition=fill_image)
        yield dict(type=UINT32, name='shape',
                   condition=fill_gradation)
        yield dict(type=BYTE, name='blur_center',
                   condition=fill_gradation)
        # 스펙 1.2의 표28 설명과는 달리,
        # 적어도 5.0.1.7에서, 있는 경우가 있고 없는 경우가 있음
        #  yield N_ARRAY(UINT32, BYTE), 'extended'  # SPEC
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
''' Extract out streams in the specified <hwp5file> to a directory.

Usage::

    hwp5proc unpack [--loglevel=<loglevel>] [--logfile=<logfile>]
                    [--vstreams | --ole]
                    <hwp5file> [<out-directory>]
    hwp5proc unpack --help

Options::

    -h --help               Show this screen
       --loglevel=<level>   Set log level.
       --logfile=<file>     Set log file.

       --vstreams           Process with virtual streams (i.e. parsed/converted
                            form of real streams)
       --ole                Treat <hwpfile> as an OLE Compound File. As a
                            result, some streams will be presented as-is. (i.e.
                            not decompressed)

Example::

    $ hwp5proc unpack samples/sample-5017.hwp
    $ ls sample-5017

Example::

    $ hwp5proc unpack --vstreams samples/sample-5017.hwp
    $ cat sample-5017/PrvText.utf8

'''
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import os.path

from .. import storage
from . import open_hwpfile


def main(args):
    filename = args['<hwp5file>']
    hwp5file = open_hwpfile(args)

    outdir = args['<out-directory>']
    if outdir is None:
        outdir, ext = os.path.splitext(os.path.basename(filename))
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    storage.unpack(hwp5file, outdir)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


def open_url(context, url):
    ''' open InputStream from a URL.

    :param url: a URL to open an InputStream.
    :returns: an instance of InputStream
    '''

    # see http://wiki.openoffice.org
    #     /wiki/Documentation/DevGuide/UCB/Using_the_UCB_API

    from hwp5.plat._uno.services import css
    css = css.bind(context)
    ucb = css.ucb.UniversalContentBroker('Local', 'Office')
    content_id = ucb.createContentIdentifier(url)
    content = ucb.queryContent(content_id)

    import unohelper
    from com.sun.star.io import XActiveDataSink

    class DataSink(unohelper.Base, XActiveDataSink):
        def setInputStream(self, stream):
            self.stream = stream

        def getInputStream(self):
            return self.stream

    datasink = DataSink()

    from com.sun.star.ucb import Command, OpenCommandArgument2
    openargs = OpenCommandArgument2()
    openargs.Mode = 2  # OpenMode.DOCUMENT
    openargs.Priority = 32768
    openargs.Sink = datasink

    command = Command()
    command.Name = 'open'
    command.Handle = -1
    command.Argument = openargs

    content.execute(command, 0, None)
    return datasink.stream



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel._shared import RecordModel
from hwp5.tagids import HWPTAG_SHAPE_COMPONENT
from hwp5.dataio import Flags
from hwp5.dataio import UINT32
from hwp5.dataio import UINT16
from hwp5.dataio import UINT8
from hwp5.dataio import SHWPUNIT
from hwp5.dataio import WORD
from hwp5.dataio import BYTE
from hwp5.dataio import X_ARRAY
from hwp5.dataio import N_ARRAY
from hwp5.dataio import Struct
from hwp5.dataio import DOUBLE
from hwp5.dataio import ref_member
from hwp5.dataio import HexBytes
from hwp5.binmodel.controlchar import CHID
from hwp5.binmodel._shared import Coord
from hwp5.binmodel._shared import BorderLine
from hwp5.binmodel._shared import FillColorPattern
from hwp5.binmodel._shared import FillGradation
from hwp5.binmodel._shared import FillImage
from hwp5.binmodel.controls.gshape_object_control import GShapeObjectControl


class Matrix(Struct):
    ''' 표 80 matrix 정보

    2D Transform Matrix

    [a c e][x]
    [b d f][y]
    [0 0 1][1]
    '''
    def attributes():
        yield DOUBLE, 'a'
        yield DOUBLE, 'c'
        yield DOUBLE, 'e'
        yield DOUBLE, 'b'
        yield DOUBLE, 'd'
        yield DOUBLE, 'f'
    attributes = staticmethod(attributes)


class ScaleRotationMatrix(Struct):
    def attributes():
        yield Matrix, 'scaler',
        yield Matrix, 'rotator',
    attributes = staticmethod(attributes)


def parent_must_be_gso(context, values):
    ''' parent record type is GShapeObjectControl '''
    # GSO-child ShapeComponent specific:
    # it may be a GSO model's attribute, e.g. 'child_chid'
    if 'parent' in context:
        parent_context, parent_model = context['parent']
        return parent_model['type'] is GShapeObjectControl


def chid_is_container(context, values):
    ''' chid == CHID.CONTAINER '''
    return values['chid'] == CHID.CONTAINER


def chid_is_rect(context, values):
    ''' chid == CHID.RECT '''
    return values['chid'] == CHID.RECT


def chid_is_rect_and_fill_colorpattern(context, values):
    ''' chid == CHID.RECT and fill_flags.fill_colorpattern '''
    return (values['chid'] == CHID.RECT and
            values['fill_flags'].fill_colorpattern)


def chid_is_rect_and_fill_image(context, values):
    ''' chid == CHID.RECT and fill_flags.fill_image '''
    return (values['chid'] == CHID.RECT and
            values['fill_flags'].fill_image)


def chid_is_rect_and_fill_gradation(context, values):
    ''' chid == CHID.RECT and fill_flags.fill_gradation '''
    return (values['chid'] == CHID.RECT and
            values['fill_flags'].fill_gradation)


def chid_is_line(context, values):
    ''' chid == CHID.LINE '''
    return values['chid'] == CHID.LINE


class ShapeComponent(RecordModel):
    ''' 4.2.9.2.1. 개체 요소 '''
    tagid = HWPTAG_SHAPE_COMPONENT
    FillFlags = Flags(UINT16,
                      8, 'fill_colorpattern',
                      9, 'fill_image',
                      10, 'fill_gradation')
    Flags = Flags(UINT32,
                  0, 'flip')

    def attributes(cls):
        ''' 표 78 개체 요소 속성 '''

        yield dict(type=CHID, name='chid0', condition=parent_must_be_gso)

        yield CHID, 'chid'
        yield SHWPUNIT, 'x_in_group'
        yield SHWPUNIT, 'y_in_group'
        yield WORD, 'level_in_group'
        yield WORD, 'local_version'
        yield SHWPUNIT, 'initial_width'
        yield SHWPUNIT, 'initial_height'
        yield SHWPUNIT, 'width'
        yield SHWPUNIT, 'height'
        yield cls.Flags, 'flags'
        yield WORD, 'angle'
        yield Coord, 'rotation_center'

        ''' 표 79 Rendering 정보 '''
        yield WORD, 'scalerotations_count'
        yield Matrix, 'translation'
        yield dict(type=X_ARRAY(ScaleRotationMatrix,
                                ref_member('scalerotations_count')),
                   name='scalerotations')

        #
        # Container
        #

        yield dict(type=N_ARRAY(WORD, CHID),
                   name='controls',
                   condition=chid_is_container)

        #
        # Rectangle
        #

        ''' 표 81 테두리 선 정보 '''
        yield dict(type=BorderLine, name='border', condition=chid_is_rect)
        ''' 표 83 Outline style '''
        # TODO: Outline ???
        yield dict(type=cls.FillFlags, name='fill_flags',
                   condition=chid_is_rect)
        yield dict(type=UINT16, name='unknown', condition=chid_is_rect)
        yield dict(type=UINT8, name='unknown1', condition=chid_is_rect)
        yield dict(type=FillColorPattern, name='fill_colorpattern',
                   condition=chid_is_rect_and_fill_colorpattern)
        yield dict(type=FillGradation, name='fill_gradation',
                   condition=chid_is_rect_and_fill_gradation)
        yield dict(type=FillImage, name='fill_image',
                   condition=chid_is_rect_and_fill_image)
        yield dict(type=UINT32, name='fill_shape',
                   condition=chid_is_rect)
        yield dict(type=BYTE, name='fill_blur_center',
                   condition=chid_is_rect_and_fill_gradation)

        # TODO: 아래 두 필드: chid == $rec일 때만인지 확인 필요
        yield dict(type=HexBytes(5), name='unknown2',
                   condition=chid_is_rect, version=(5, 0, 2, 4))
        yield dict(type=HexBytes(16), name='unknown3',
                   condition=chid_is_rect, version=(5, 0, 2, 4))

        #
        # Line
        #

        yield dict(type=BorderLine, name='line',
                   condition=chid_is_line)
    attributes = classmethod(attributes)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from hwp5.binmodel.controls.bookmark_control import BookmarkControl
from hwp5.binmodel.controls.columns_def import ColumnsDef
from hwp5.binmodel.controls.common_controls import CommonControl
from hwp5.binmodel.controls.dutmal import Dutmal
from hwp5.binmodel.controls.field import Field
from hwp5.binmodel.controls.field import FieldUnknown
from hwp5.binmodel.controls.field import FieldDate
from hwp5.binmodel.controls.field import FieldDocDate
from hwp5.binmodel.controls.field import FieldPath
from hwp5.binmodel.controls.field import FieldBookmark
from hwp5.binmodel.controls.field import FieldMailMerge
from hwp5.binmodel.controls.field import FieldCrossRef
from hwp5.binmodel.controls.field import FieldFormula
from hwp5.binmodel.controls.field import FieldClickHere
from hwp5.binmodel.controls.field import FieldClickHereData
from hwp5.binmodel.controls.field import FieldSummary
from hwp5.binmodel.controls.field import FieldUserInfo
from hwp5.binmodel.controls.field import FieldHyperLink
from hwp5.binmodel.controls.field import FieldMemo
from hwp5.binmodel.controls.field import FieldPrivateInfoSecurity
from hwp5.binmodel.controls.gshape_object_control import GShapeObjectControl
from hwp5.binmodel.controls.header_footer import HeaderFooter
from hwp5.binmodel.controls.hidden_comment import HiddenComment
from hwp5.binmodel.controls.index_marker import IndexMarker
from hwp5.binmodel.controls.note import Note
from hwp5.binmodel.controls.note import FootNote
from hwp5.binmodel.controls.note import EndNote
from hwp5.binmodel.controls.numbering import AutoNumbering
from hwp5.binmodel.controls.numbering import NewNumbering
from hwp5.binmodel.controls.page_hide import PageHide
from hwp5.binmodel.controls.page_number_position import PageNumberPosition
from hwp5.binmodel.controls.page_odd_even import PageOddEven
from hwp5.binmodel.controls.section_def import SectionDef
from hwp5.binmodel.controls.table_control import TableControl
from hwp5.binmodel.controls.tcps_control import TCPSControl

# suppress pyflake8 warning 'imported but not used'
BookmarkControl
ColumnsDef
CommonControl
Dutmal
Field
FieldUnknown
FieldDate
FieldDocDate
FieldPath
FieldBookmark
FieldMailMerge
FieldCrossRef
FieldFormula
FieldClickHere
FieldClickHereData
FieldSummary
FieldUserInfo
FieldHyperLink
FieldMemo
FieldPrivateInfoSecurity
GShapeObjectControl
HeaderFooter
HiddenComment
IndexMarker
Note
FootNote
EndNote
AutoNumbering
NewNumbering
PageHide
PageNumberPosition
PageOddEven
SectionDef
TableControl
TCPSControl



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals


def get_unichr_lang(uch):
    # Hangul Syllables
    # U+AC00..U+D7AF
    # Hangul Jamo Extended-B
    # U+D7B0..D7FF
    if u'\uAC00' <= uch <= u'\uD7FF':
        return 'ko'

    # Control Characters and Numbers in Basic Latin
    if u'\u0000' <= uch <= u'\u0040':
        return None

    # Hangul Jamo
    if u'\u1100' <= uch <= u'\u11FF':
        return 'ko'

    # Hangul Compatibility Jamo
    if u'\u3130' <= uch <= u'\u318F':
        return 'ko'

    # Hangul Jamo Extended-A
    if u'\uA960' <= uch <= u'\uA97F':
        return 'ko'

    # -- en --

    # Basic Latin, Latin Extended-A/B
    if u'\u0040' <= uch <= u'\u024F':
        return 'en'

    # -- cn --

    # CJK Unified Ideographs
    # U+4E00..U+9FFF
    if u'\u4E00' <= uch <= u'\u9FFF':
        return 'cn'

    # CJK Radicals Supplement
    # U+2E80..U+2EFF
    # Kangxi Radicals
    # U+2F00..U+2FDF
    if u'\u2E80' <= uch <= u'\u2FDF':
        return 'cn'

    # CJK Unified Ideographs Extension A
    # U+3400..U+4DBF
    if u'\u3400' <= uch <= u'\u4DBF':
        return 'cn'

    # CJK Compatibility Ideographs
    # U+F900..U+FAFF
    if u'\uF900' <= uch <= u'\uFAFF':
        return 'cn'

    # CJK Symbols and Punctuation
    # U+3000..U+303F
    if u'\u3000' <= uch <= u'\u303F':
        return 'symbol'

    # -- jp --

    # Hiragana + Katakana
    if u'\u3040' <= uch <= u'\u30FF':
        return 'jp'

    return 'other'


def tokenize_unicode_by_lang(text):
    buf = []
    buf_lang = None
    for uch in text:
        lang = get_unichr_lang(uch)
        if lang is None:
            buf.append(uch)
            continue
        if buf_lang == lang or buf_lang is None:
            buf_lang = lang
            buf.append(uch)
            continue
        else:
            yield buf_lang or 'ko', ''.join(buf)
            buf = [uch]
            buf_lang = lang
    if buf:
        yield buf_lang or 'ko', ''.join(buf)



# -*- coding: utf-8 -*-
#
#   pyhwp : hwp file format parser in python
#   Copyright (C) 2010-2015 mete0r <mete0r@sarangbang.or.kr>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from contextlib import contextmanager
from subprocess import CalledProcessError
import logging
import subprocess


logger = logging.getLogger(__name__)

executable = 'xmllint'
enabled = None


def xmllint_reachable():
    args = [executable, '--version']
    try:
        subprocess.check_output(args)
    except OSError:
        return False
    except CalledProcessError:
        return False
    except Exception as e:
        logger.exception(e)
        return False
    else:
        return True


def is_enabled():
    global enabled
    if enabled is None:
        enabled = xmllint_reachable()
    return enabled


def enable():
    global enabled
    enabled = True


def disable():
    global enabled
    enabled = False


def relaxng(rng_path, inp_path):
    from subprocess import Popen
    args = [executable, '--noout', '--relaxng', rng_path, inp_path]
    p = Popen(args)
    p.wait()
    return p.returncode == 0


def relaxng_compile(rng_path):
    return RelaxNG(rng_path)


class RelaxNG:

    def __init__(self, rng_path):
        self.rng_path = rng_path

    @contextmanager
    def validating_output(self, output):
        args = [executable, '--relaxng', self.rng_path, '-']
        p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=output)
        try:
            yield p.stdin
        except:
            p.stdin.close()
            p.wait()
            raise
        else:
            p.stdin.close()
            p.wait()
            if p.returncode != 0:
                raise Exception('RelaxNG validation failed')
